"""
Extract specific statements from a Wikidata JSON file.
"""
import typing
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Extract specific statements from a Wikidata JSON file.'
    }


def add_arguments(parser: KGTKArgumentParser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.io.kgtkreader import KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter

    parser.add_input_file()
    parser.add_output_file()
    
    parser.add_argument(
        "--find",
        action="store",
        type=str,
        dest="entity_ids",
        default=None,
        nargs="+",
        help='The entity IDs to find.')

    parser.add_argument(
        "--input-limit", "--limit",
        action="store",
        type=int,
        dest="input_limit",
        default=None,
        help='number of lines of input file to run on, default runs on all')

    parser.add_argument(
        "--output-limit",
        action="store",
        type=int,
        dest="output_limit",
        default=None,
        help='Limit the number of lines of output (default unlimited)')

    parser.add_argument(
        "--use-mgzip-for-input",
        nargs='?',
        type=optional_bool,
        dest="use_mgzip_for_input",
        const=True,
        default=False,
        metavar="True/False",
        help="If true, use the multithreaded gzip package, mgzip, for input. (default=%(default)s).")

    parser.add_argument(
        "--use-mgzip-for-output",
        nargs='?',
        type=optional_bool,
        dest="use_mgzip_for_output",
        const=True,
        default=False,
        metavar="True/False",
        help="If true, use the multithreaded gzip package, mgzip, for output. (default=%(default)s).")

    parser.add_argument(
        "--mgzip-threads-for-input",
        type=int,
        default=KgtkReaderOptions.MGZIP_THREAD_COUNT_DEFAULT,
        dest="mgzip_threads_for_input",
        help="The number of threads per mgzip input streama. (default=%(default)s).")

    parser.add_argument(
        "--mgzip-threads-for-output",
        type=int,
        default=KgtkWriter.MGZIP_THREAD_COUNT_DEFAULT,
        dest="mgzip_threads_for_output",
        help="The number of threads per mgzip output streama. (default=%(default)s).")

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        entity_ids: typing.List[str],
        input_limit: int,
        output_limit: int,
        use_mgzip_for_input: bool,
        use_mgzip_for_output: bool,
        mgzip_threads_for_input: int,
        mgzip_threads_for_output: int,
        ):

    import simplejson as json
    import sys

    in_path = KGTKArgumentParser.get_input_file(input_file)
    out_path = KGTKArgumentParser.get_output_file(output_file)

    from gzip import GzipFile
    print("Processing.", file=sys.stderr, flush=True)

    # Open the input file first to make it easier to monitor with "pv".
    input_f: typing.Union[GzipFile, typing.IO[typing.Any]]
    if str(in_path) == "-":
        print('Processing wikidata from standard input', file=sys.stderr, flush=True)
        # It is not well documented, but this is how you read binary data
        # from stdin in Python 3.
        input_f = sys.stdin.buffer

    else:
        print('Processing wikidata file %s' % str(in_path), file=sys.stderr, flush=True)
        input_f = open(in_path, mode='rb')
            
        if str(in_path).endswith(".bz2"):
            import bz2
            print('Decompressing (bz2)', file=sys.stderr, flush=True)
            # TODO: Optionally use a system decompression program.
            input_f = bz2.open(input_f)

        elif str(in_path).endswith(".gz"):
            # TODO: Optionally use a system decompression program.
            if use_mgzip_for_input:
                import mgzip
                print('Decompressing (mgzip)', file=sys.stderr, flush=True)
                input_f = mgzip.open(input_f, thread=mgzip_threads_for_input)
            else:
                import gzip
                print('Decompressing (gzip)', file=sys.stderr, flush=True)
                input_f = gzip.open(input_f)

    # Open the input file first to make it easier to monitor with "pv".
    output_f: typing.Union[GzipFile, typing.IO[typing.Any]]
    if str(out_path) == "-":
        print('Sending wikidata JSON to standatd output', file=sys.stderr, flush=True)
        # It is not well documented, but this is how you write binary data
        # from stdin in Python 3.
        output_f = sys.stdout.buffer

    else:
        print('Writing wikidata file %s' % str(out_path), file=sys.stderr, flush=True)
        output_f = open(out_path, mode='wb')
            
        if str(out_path).endswith(".bz2"):
            import bz2
            print('Compressing (bz2)', file=sys.stderr, flush=True)
            # TODO: Optionally use a system decompression program.
            output_f = bz2.open(output_f, "wb")

        elif str(out_path).endswith(".gz"):
            # TODO: Optionally use a compression program.
            if use_mgzip_for_output:
                import mgzip
                print('Compressing (mgzip)', file=sys.stderr, flush=True)
                output_f = mgzip.open(output_f, "wb", thread=mgzip_threads_for_output)
            else:
                import gzip
                print('Compressing (gzip)', file=sys.stderr, flush=True)
                output_f = gzip.open(output_f, "wb")

    entity_id_set: typing.Set[str] = set(entity_ids)

    output_count: int = 0
    input_count: int
    line: bytes
    for input_count, line in enumerate(input_f):
        if input_limit and input_count >= input_limit:
            break
        clean_line = line.strip()
        if clean_line.endswith(b","):
            clean_line = clean_line[:-1]
        if len(clean_line) > 1:
            obj = json.loads(clean_line)
            entity = obj["id"]
            if entity in entity_id_set:
                if output_count == 0:
                    output_f.write(b"[\n")
                else:
                    output_f.write(b",\n")
                output_f.write(clean_line)
                output_count += 1
                if output_limit is not None and output_count >= output_limit:
                    break
            
    print('Done processing {}'.format(str(in_path)), file=sys.stderr, flush=True)
    input_f.close()

    if output_count > 0:
        output_f.write(b"\n]\n")
    output_f.close()
   
    print('Wrote {} records'.format(output_count), file=sys.stderr, flush=True)



"""Validate language qualifiers.

TODO: Extend our language tag recognition to RFC 3066.
https://tools.ietf.org/html/rfc3066

TODO: Cite RDF compliance:  BCP46 Section 2-2-9.
https://tools.ietf.org/html/bcp47#section-2.2.9
https://www.w3.org/TR/2014/REC-rdf11-concepts-20140225/#section-Graph-Literal

TODO: For tull Wikimedia compatability, consider supporting Wimikedia's additional
codes.  See the following references:

https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all
https://meta.wikimedia.org/wiki/Special_language_codes

"""

from argparse import ArgumentParser, Namespace
import attr
import iso639 # type: ignore
import pycountry # type: ignore
import typing

from kgtk.value.kgtkvalueoptions import KgtkValueOptions, DEFAULT_KGTK_VALUE_OPTIONS

# Problem: pycountry incorporates the Debian team's ISO 639-3 table,
# which as of 03-May-2020 has not been updated in four years!
# Meanwhile, iso639 (from pypi iso-639) has an ISO 639-3 table
# from 2015-05-05.
#
# https://salsa.debian.org/iso-codes-team/iso-codes/-/blob/master/iso_639-3/iso_639_3.tab 
# https://pypi.org/project/iso-639/
#
# Problem: Wikidata may contain obsolete language codes which have been
# removed from the standard indices.
#
# Example: "mo"
#
# Solution: We will keep a list of additional language codes.
@attr.s(slots=True, frozen=True)
class LanguageValidator:
    """
    The language code may be a two- or three-character code from ISO
    639-3, which replaces ISO 639-1 and ISO 639-2.  In addition, wikidata
    may include language codes, such as 'mo', that have been retired.  The
    additional_language_codes table supports these codes, when allowed.

    Wikidata may also contain collective language codes, such as "nah",
    referring the the Nahuatl languages. These codes from ISO 639-5 are
    accepted as a fallback when ISO 639-3 lookup fails.

    https://meta.wikimedia.org/wiki/Special_language_codes
    https://en.wikipedia.org/wiki/Template:ISO_639_name_be-tarask
    """

    DEFAULT_ADDITIONAL_LANGUAGE_CODES: typing.List[str] = [
        # New codes:
        "cnr", # Montenegrin.  Added 21-Dec-2017. https://iso639-3.sil.org/code/cnr
        "hyw", # Western Armenian.  Added 23-Jan-2018. https://iso639-3.sil.org/code/hyw
        "szy", # Sakizawa.  Added 25-Jan-2019. https://iso639-3.sil.org/code/szy

        # Obsolete codes:
        "bh", # Bihari lanuages, apparently replaced by "bih".
        "mo", # Moldavian. Retired 3-Nov-2008. Replaced by the codes for Romanian.
              # http://www.personal.psu.edu/ejp10/blogs/gotunicode/2008/11/language-tage-mo-for-moldovan.html
        "eml", # Emiliano-Romagnolo. Split and retired 16-Jan-2009. https://iso639-3.sil.org/code/eml
    ]

    @classmethod
    def validate(cls,
                 lang: str,
                 options: KgtkValueOptions=DEFAULT_KGTK_VALUE_OPTIONS,
                 verbose: bool = False,
    )->bool:
        # Wikidata contains entries such as:
        # 'panamenha'@pt-br      # language code followed by country code
        # 'Ecuador'@es-formal    # language code followed by dialect name
        #
        # If we see a dash, we'll check the language code by itself.
        if verbose:
            print("Validating '%s'" % lang)

        save_lang: str = lang # for the debug prints below.
        country_or_dialect: str = ""
        if options.allow_language_suffixes and "-" in lang:
            (lang, country_or_dialect) = lang.split("-", 1)
            if verbose:
                print("'%s' split into '%s' and '%s'" % (save_lang, lang, country_or_dialect))

        if len(lang) == 2:
            # Two-character language codes.
            if pycountry.languages.get(alpha_2=lang) is not None:
                if verbose:
                    print("pycountry.languages.get(alpha_2=lang) succeeded")
                return True

        elif len(lang) == 3:
            # Three-character language codes.
            if pycountry.languages.get(alpha_3=lang) is not None:
                if verbose:
                    print("pycountry.languages.get(alpha_3=lang) succeeded")
                return True

        # Perhaps this is a collective (language family) code from ISO 639-5?
        try:
            iso639.languages.get(part5=lang)
            if verbose:
                print("iso639.languages.get(part5=lang) succeeded")
            return True
        except KeyError:
            pass

        # If there's a table of additional language codes, check there:
        additional_language_codes: typing.List[str]
        if options.additional_language_codes is not None:
            additional_language_codes = options.additional_language_codes
            if verbose:
                print("Using a custom list of %d additional language codes." % len(additional_language_codes))
        else:
            if verbose:
                print("Using the default list of additional language codes.")
            additional_language_codes = LanguageValidator.DEFAULT_ADDITIONAL_LANGUAGE_CODES

        if lang in additional_language_codes:
            if verbose:
                print("found in the table of additional languages.")
            return True

        if verbose:
            print("Not found.")
        return False

def main():
    """
    Test the language validator.
    """
    parser: ArgumentParser = ArgumentParser()
    parser.add_argument(dest="values", help="The values(s) to test", type=str, nargs="+")
    parser.add_argument("-v", "--verbose", dest="verbose", help="Print additional progress messages.", action='store_true')
    KgtkValueOptions.add_arguments(parser)
    args: Namespace = parser.parse_args()

    # Build the value parsing option structure.
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

    value: str
    for value in args.values:
        result: bool = LanguageValidator.validate(value, options=value_options, verbose=args.verbose)                                   
        print("%s: %s" % (value, str(result)), flush=True)

if __name__ == "__main__":
    main()



"""
Export wikidata, mirroring import_wikidata.
"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Export wikidata from a set of KGTK files.',
        'description': 'Export wikidata, mirroring import-wikidata. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert lift --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(who="A KGTK file with node records",
                          dest="node_file",
                          options=["--node-file"],
                          optional=False,
                          default_stdin=False)
    parser.add_input_file(who="A KGTK file with edge records",
                          dest="edge_file",
                          options=["--edge-file"],
                          optional=False,
                          default_stdin=False)
    parser.add_input_file(who="A KGTK file with qualifier records",
                          dest="qualifier_file",
                          options=["--qualifier-file"],
                          optional=False,
                          default_stdin=False)
    parser.add_output_file()


    KgtkReader.add_debug_arguments(parser, expert=_expert)
    # TODO: seperate reader_options for the label file.
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(node_file: KGTKFiles,
        edge_file: KGTKFiles,
        qualifier_file: KGTKFiles,
        output_file: KGTKFiles,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException
    from kgtk.exports.exportwikidata import ExportWikidata
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    node_kgtk_file: Path = KGTKArgumentParser.get_input_file(node_file, who="KGTK node file", default_stdin=False)
    edge_kgtk_file: Path = KGTKArgumentParser.get_input_file(edge_file, who="KGTK edge file", default_stdin=False)
    qualifier_kgtk_file: Path = KGTKArgumentParser.get_input_file(qualifier_file, who="KGTK qualifier file", default_stdin=False)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--node-file=%s" % str(node_kgtk_file), file=error_file, flush=True)
        print("--edge-file=%s" % str(edge_kgtk_file), file=error_file, flush=True)
        print("--qualifier-file=%s" % str(qualifier_kgtk_file), file=error_file, flush=True)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file, flush=True)

        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        ew: ExportWikidata = ExportWikidata(
            node_file_path=node_kgtk_file,
            edge_file_path=edge_kgtk_file,
            qualifier_file_path=qualifier_kgtk_file,
            output_file_path=output_kgtk_file,

            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        ew.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




import unittest
import os
from kgtk.generator import TripleGenerator
from pathlib import Path


class TestTripleGeneration(unittest.TestCase):
    def test_truthy_dates_generation(self):
        dates_tsv_file = Path('data/dates.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        o = 'data/dates_truthy_tmp.ttl'
        generator = TripleGenerator(prop_file=wikidata_property_file, label_set='label', alias_set='aliases',
                                    description_set='descriptions', warning=True, n=100, truthy=True, use_id=True,
                                    dest_fp=o, log_path="data/date_warning.log", prop_declaration=False,
                                    prefix_path="NONE", input_file=dates_tsv_file, error_action='log')
        generator.process()

        f1 = open('data/dates_truthy.ttl')
        f2 = open('data/dates_truthy_tmp.ttl')

        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()
        self.assertEqual(os.stat("data/date_warning.log").st_size, 0)

        p = Path("data/date_warning.log")
        p.unlink()
        p = Path('data/dates_truthy_tmp.ttl')
        p.unlink()

    def test_truthy_property_triple_generation(self):
        property_tsv_file = Path('data/P10.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        o = 'data/P10_truthy_tmp.ttl'
        generator = TripleGenerator(prop_file=wikidata_property_file, label_set='label', alias_set='aliases',
                                    description_set='descriptions', warning=True, n=100, truthy=True, use_id=True,
                                    dest_fp=o, log_path="data/warning.log", prop_declaration=False, prefix_path="NONE",
                                    input_file=property_tsv_file, error_action='log')
        generator.process()

        f1 = open('data/P10_truthy.ttl')
        f2 = open('data/P10_truthy_tmp.ttl')
        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()

        self.assertEqual(os.stat("data/warning.log").st_size, 0)
        p = Path("data/warning.log")
        p.unlink()
        p = Path('data/P10_truthy_tmp.ttl')
        p.unlink()

    def test_property_triple_generation(self):
        property_tsv_file = Path('data/P10.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        o = 'data/P10_not_truthy_tmp.ttl'
        generator = TripleGenerator(prop_file=wikidata_property_file, label_set='label', alias_set='aliases',
                                    description_set='descriptions', warning=True, n=100, truthy=False, use_id=True,
                                    dest_fp=o, log_path="data/warning.log", prop_declaration=False, prefix_path="NONE",
                                    input_file=property_tsv_file, error_action='log')
        generator.process()

        f1 = open('data/P10_not_truthy.ttl')
        f2 = open('data/P10_not_truthy_tmp.ttl')
        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()
        self.assertEqual(os.stat("data/warning.log").st_size, 0)
        p = Path("data/warning.log")
        p.unlink()
        p = Path('data/P10_not_truthy_tmp.ttl')
        p.unlink()

    def test_truthy_qnode_triple_generation(self):
        qnode_tsv_file = Path('data/Q57160439.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        o = 'data/Q57160439_truthy_tmp.ttl'
        generator = TripleGenerator(prop_file=wikidata_property_file, label_set='label', alias_set='aliases',
                                    description_set='descriptions', warning=True, n=100, truthy=True, use_id=True,
                                    dest_fp=o, log_path="data/warning.log", prop_declaration=False, prefix_path="NONE",
                                    input_file=qnode_tsv_file, error_action='log')
        generator.process()

        f1 = open('data/Q57160439_truthy.ttl')
        f2 = open('data/Q57160439_truthy_tmp.ttl')
        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()
        self.assertEqual(os.stat("data/warning.log").st_size, 0)
        p = Path("data/warning.log")
        p.unlink()
        p = Path('data/Q57160439_truthy_tmp.ttl')
        p.unlink()

    def test_not_truthy_qnode_triple_generation(self):
        qnode_tsv_file = Path('data/Q57160439.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        o = 'data/Q57160439_not_truthy_tmp.ttl'
        generator = TripleGenerator(prop_file=wikidata_property_file, label_set='label', alias_set='aliases',
                                    description_set='descriptions', warning=True, n=100, truthy=False, use_id=True,
                                    dest_fp=o, log_path="data/warning.log", prop_declaration=False, prefix_path="NONE",
                                    input_file=qnode_tsv_file, error_action='log')
        generator.process()

        f1 = open('data/Q57160439_not_truthy.ttl')
        f2 = open('data/Q57160439_not_truthy_tmp.ttl')

        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()

        self.assertEqual(os.stat("data/warning.log").st_size, 0)

        p = Path("data/warning.log")
        p.unlink()
        p = Path('data/Q57160439_not_truthy_tmp.ttl')
        p.unlink()

    def test_triple_small_values(self):
        small_values_file = Path('data/small_values.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        o = 'data/small_values_tmp.ttl'
        generator = TripleGenerator(prop_file=wikidata_property_file, label_set='label', alias_set='aliases',
                                    description_set='descriptions', warning=True, n=100, truthy=True, use_id=True,
                                    dest_fp=o, log_path="data/warning.log", prop_declaration=False, prefix_path="NONE",
                                    input_file=small_values_file, error_action='log')
        generator.process()

        f1 = open('data/small_values.ttl')
        f2 = open('data/small_values_tmp.ttl')
        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()
        self.assertEqual(os.stat("data/warning.log").st_size, 0)
        p = Path("data/warning.log")
        p.unlink()
        p = Path('data/small_values_tmp.ttl')
        p.unlink()

    def test_triple_corrupted_edges(self):
        corrupted_kgtk_file = Path('data/corrupted_kgtk.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        o = 'data/corrupted_tmp.ttl'
        generator = TripleGenerator(prop_file=wikidata_property_file, label_set='label', alias_set='aliases',
                                    description_set='descriptions', warning=True, n=100, truthy=True, use_id=True,
                                    dest_fp=o, log_path="data/corrupted_warning_tmp.log", prop_declaration=False,
                                    prefix_path="NONE", input_file=corrupted_kgtk_file, error_action='log')
        generator.process()

        f1 = open('data/corrupted.ttl')
        f2 = open('data/corrupted_tmp.ttl')
        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()
        f1 = open("data/corrupted_warning.log")
        f2 = open("data/corrupted_warning_tmp.log")
        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()
        p = Path("data/corrupted_warning_tmp.log")
        p.unlink()
        p = Path('data/corrupted_tmp.ttl')
        p.unlink()



from distutils.core import setup
from setuptools import find_packages  # type: ignore
from setuptools.command.build_py import build_py
from kgtk import __version__
import sys

lite_build = False
if '--lite' in sys.argv:
    lite_build = True
    sys.argv.remove('--lite')

with open('requirements.txt', 'r') as f:
    install_requires = list()
    dependency_links = list()
    for line in f:
        re = line.strip()
        if re:
            if re.startswith('git+') or re.startswith('svn+') or re.startswith('hg+'):
                dependency_links.append(re)
            else:
                install_requires.append(re)

if not lite_build:
    with open('requirements-full.txt', 'r') as f:
        for line in f:
            re = line.strip()
            if re:
                if re.startswith('git+') or re.startswith('svn+') or re.startswith('hg+'):
                    dependency_links.append(re)
                else:
                    install_requires.append(re)

lite_excluded_modules = {
    'kgtk.cli': {'filter', 'export_gt', 'export_neo4j', 'connected-components', 'graph_statistics', 'gt_loader',
                 'reachable_nodes', 'text_embedding'}
}


class kgtk_build_py(build_py):

    def find_package_modules(self, package, package_dir):
        modules = super().find_package_modules(package, package_dir)
        if lite_build and package in lite_excluded_modules:
            modules = list(filter(lambda m: m[1] not in lite_excluded_modules[package], modules))
        return modules


packages = find_packages()

setup(
    name='kgtk' if not lite_build else 'kgtk-lite',
    version=__version__,
    packages=packages,
    url='https://github.com/usc-isi-i2/kgtk',
    license='MIT',
    author='ISI CKGs',
    include_package_data=True,
    install_requires=install_requires,
    dependency_links=dependency_links,
    entry_points={
        'console_scripts': [
            'kgtk = kgtk.cli_entry:cli_entry',
        ],
    },
    cmdclass={
        'build_py': kgtk_build_py
    }
)



"""Import ntriples into KGTK format.
"""
from argparse import ArgumentParser, Namespace
import attr
import csv
from pathlib import Path
import re
import shortuuid # type: ignore
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalue import KgtkValue, KgtkValueFields
from kgtk.value.kgtkvalueoptions import KgtkValueOptions, DEFAULT_KGTK_VALUE_OPTIONS

@attr.s(slots=True, frozen=False)
class KgtkNtriples(KgtkFormat):
    # Class attributes:
    DEFAULT_PREFIX_EXPANSION: str = "prefix_expansion"
    DEFAULT_STRUCTURED_VALUE: str = "kgtk:structured_value"
    DEFAULT_STRUCTURED_URI: str = "kgtk:structured_uri"
    DEFAULT_NAMESPACE_ID_PREFIX: str = "n"
    DEFAULT_NAMESPACE_ID_COUNTER: int = 1
    DEFAULT_NAMESPACE_ID_ZFILL: int = 0
    DEFAULT_NAMESPACE_ID_USE_UUID: bool = False
    DEFAULT_OUTPUT_ONLY_USED_NAMESPACES: bool = True
    DEFAULT_NEWNODE_PREFIX: str = "kgtk:node"
    DEFAULT_NEWNODE_COUNTER: int = 1
    DEFAULT_NEWNODE_ZFILL: int = 0
    DEFAULT_NEWNODE_USE_UUID: bool = False
    DEFAULT_LOCAL_NAMESPACE_PREFIX: str = "X"
    DEFAULT_LOCAL_NAMESPACE_USE_UUID: bool = True
    DEFAULT_ALLOW_LAX_URI: bool = True
    DEFAULT_BUILD_ID: bool = False
    DEFAULT_ESCAPE_PIPES: bool = True
    DEFAULT_VALIDATE: bool = False

    COLUMN_NAMES: typing.List[str] = [KgtkFormat.NODE1, KgtkFormat.LABEL, KgtkFormat.NODE2]
    
    # The following lexical analysis is based on:
    # https://docs.python.org/3/reference/lexical_analysis.html

    # The long integer suffix was part of Python 2.  It was dropped in Python 3.
    long_suffix_pat: str = r'[lL]'

    plus_or_minus_pat: str = r'[-+]'

    # Integer literals.
    #
    # Decimal integers, allowing leading zeros.
    digit_pat: str = r'[0-9]'
    decinteger_pat: str = r'(?:{digit}(?:_?{digit})*{long_suffix}?)'.format(digit=digit_pat,
                                                                            long_suffix=long_suffix_pat)
    bindigit_pat: str = r'[01]'
    bininteger_pat: str = r'(?:0[bB](":_?{bindigit})+{long_suffix})'.format(bindigit=bindigit_pat,
                                                                            long_suffix=long_suffix_pat)
    octdigit_pat: str = r'[0-7]'
    octinteger_pat: str = r'(?:0[oO](":_?{octdigit})+{long_suffix})'.format(octdigit=octdigit_pat,
                                                                            long_suffix=long_suffix_pat)
    hexdigit_pat: str = r'[0-7a-fA-F]'
    hexinteger_pat: str = r'(?:0[xX](":_?{hexdigit})+{long_suffix})'.format(hexdigit=hexdigit_pat,
                                                                            long_suffix=long_suffix_pat)
     
    integer_pat: str = r'(?:{decinteger}|{bininteger}|{octinteger}|{hexinteger})'.format(decinteger=decinteger_pat,
                                                                                         bininteger=bininteger_pat,
                                                                                         octinteger=octinteger_pat,
                                                                                         hexinteger=hexinteger_pat)

    # Floating point literals.
    digitpart_pat: str = r'(?:{digit}(?:_?{digit})*)'.format(digit=digit_pat)
    fraction_pat: str = r'(?:\.{digitpart})'.format(digitpart=digitpart_pat)
    pointfloat_pat: str = r'(?:{digitpart}?{fraction})|(?:{digitpart}\.)'.format(digitpart=digitpart_pat,
                                                                                 fraction=fraction_pat)
    exponent_pat: str = r'(?:[eE]{plus_or_minus}?{digitpart})'.format(plus_or_minus=plus_or_minus_pat,
                                                                      digitpart=digitpart_pat)
    exponentfloat_pat: str = r'(?:{digitpart}|{pointfloat}){exponent}'.format(digitpart=digitpart_pat,
                                                                              pointfloat=pointfloat_pat,
                                                                              exponent=exponent_pat)
    floatnumber_pat: str = r'(?:{pointfloat}|{exponentfloat})'.format(pointfloat=pointfloat_pat,
                                                                      exponentfloat=exponentfloat_pat)

    # Imaginary literals.
    imagnumber_pat: str = r'(?:{floatnumber}|{digitpart})[jJ]'.format(floatnumber=floatnumber_pat,
                                                                      digitpart=digitpart_pat)

    # Numeric literals.
    numeric_pat: str = r'(?:{plus_or_minus}?(?:{integer}|{floatnumber}|{imagnumber}))'.format(plus_or_minus=plus_or_minus_pat,
                                                                                              integer=integer_pat,
                                                                                              floatnumber=floatnumber_pat,
                                                                                              imagnumber=imagnumber_pat)

    # A URI must begin with a scheme per RFC 3986.
    #
    # We don't use this because ISI was inserting some invalid URIs that did
    # not include a scheme, but kept the definition for reference.
    SCHEME_PAT: str = r'^(?:[a-zA_Z][-.+0-9a-zA_Z]*://)'

    # The URI is delimited by matching angle brackets.  Angle brackets
    # cannot appear in a valid URI per RFC 3986.
    URI_PAT: str = r'(?:<[^>]+>)'

    # This is a guess about what may be in a blank node. It's entirely
    # possible that other characters, sych as hyphen, might be allowed.
    BLANK_NODE_PAT: str = r'(?:_:[0-9a-zA-Z_]+)'

    # Double quoted strings with backslash escapes.
    STRING_PAT: str = r'"(?:[^\\]|(?:\\.))*"'

    STRUCTURED_VALUE_PAT: str = r'(?:{string}(?:\^\^{uri})?)'.format(string=STRING_PAT, uri=URI_PAT)
    FIELD_PAT: str = r'(?:{uri}|{blank_node}|{structured_value}|{numeric})'.format(uri=URI_PAT,
                                                                                   blank_node=BLANK_NODE_PAT,
                                                                                   structured_value=STRUCTURED_VALUE_PAT,
                                                                                   numeric=numeric_pat,
    )
    ROW_PAT: str = r'(?P<node1>{field})\s(?P<label>{field})\s(?P<node2>{field})\s\.'.format(field=FIELD_PAT)
    ROW_RE: typing.Pattern = re.compile(r'^' + ROW_PAT + r'$')

    SLASH_HASH_PAT: str = r'/|#'
    SLASH_HASH_RE: typing.Pattern = re.compile(SLASH_HASH_PAT)

    # Instance attributes:

    # TODO: write a validator:
    input_file_paths: typing.List[Path] = attr.ib()

    output_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))

    reject_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    namespace_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    updated_namespace_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    reader_options: KgtkReaderOptions = attr.ib(validator=attr.validators.instance_of(KgtkReaderOptions))

    allow_lax_uri: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_LAX_URI)

    local_namespace_prefix: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_LOCAL_NAMESPACE_PREFIX)
    local_namespace_use_uuid: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_LOCAL_NAMESPACE_USE_UUID)

    namespace_id_prefix: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_NAMESPACE_ID_PREFIX)
    namespace_id_counter: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_NAMESPACE_ID_COUNTER)
    namespace_id_zfill: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_NAMESPACE_ID_ZFILL)
    namespace_id_use_uuid: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_NAMESPACE_ID_USE_UUID)
    output_only_used_namespaces: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_LOCAL_NAMESPACE_USE_UUID)

    prefix_expansion_label: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_PREFIX_EXPANSION)

    structured_value_label: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_STRUCTURED_VALUE)
    structured_uri_label: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_STRUCTURED_URI)

    newnode_prefix: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_NEWNODE_PREFIX)
    newnode_counter: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_NEWNODE_COUNTER)
    newnode_zfill: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_NEWNODE_ZFILL)
    newnode_use_uuid: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_NEWNODE_USE_UUID)

    build_id: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_BUILD_ID)
    idbuilder_options: typing.Optional[KgtkIdBuilderOptions] = attr.ib(default=None)
    idbuilder: typing.Optional[KgtkIdBuilder] = attr.ib(default=None)

    validate: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_VALIDATE)
    value_options: KgtkValueOptions = attr.ib(validator=attr.validators.instance_of(KgtkValueOptions), default=DEFAULT_KGTK_VALUE_OPTIONS)

    override_uuid: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    local_namespace_uuid: str = attr.ib(default="")

    namespace_prefixes: typing.MutableMapping[str, str] = attr.ib(factory=dict)
    namespace_ids: typing.MutableMapping[str, str] = attr.ib(factory=dict)
    used_namespaces: typing.Set[str] = attr.ib(factory=set)

    escape_pipes: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ESCAPE_PIPES)

    output_line_count: int = attr.ib(default=0)

    def write_row(self, ew: KgtkWriter, node1: str, label: str, node2: str):
        output_row: typing.List[str] = [ node1, label, node2]
        if self.idbuilder is None:
            ew.write(output_row)
        else:
            ew.write(self.idbuilder.build(output_row, self.output_line_count))
        self.output_line_count += 1

    def convert_blank_node(self, item: str)->typing.Tuple[str, bool]:
        body: str = item[1:] # Strip the leading underscore, keep the colon.
        if self.local_namespace_use_uuid:
            return self.local_namespace_prefix + self.local_namespace_uuid + body, True
        else:
            return self.local_namespace_prefix + body, True

    def convert_uri(self, item: str, line_number: int)->typing.Tuple[str, bool]:
        body: str = item[1:-1] # Strip off the enclosing brackets.

        # First, check for an exact match for the body:
        namespace_id: str
        if body in self.namespace_prefixes:
            # Yes, this prefix exactly matches the body.
            namespace_id = self.namespace_prefixes[body]

            if self.output_only_used_namespaces:
                self.used_namespaces.add(namespace_id)

            # Return the namespace-prefixed URI:
            return namespace_id + ":", True

        # Move past "http://" or "https://":
        after_slashslash: int
        slashslash: int =  body.rfind("://")
        if slashslash < 1:
            if not self.allow_lax_uri:
                if self.verbose:
                    print("Line %d: invalid URI: '%s'" % (line_number, item), file=self.error_file, flush=True)
                return item, False
            after_slashslash = 0
        else:
            after_slashslash = slashslash + len("://")

        # Build a left-to-right list of slash- or hash-terminated sections:
        matches: typing.List[int] = [ ]
        m: typing.Match
        for m in self.SLASH_HASH_RE.finditer(body, after_slashslash):
            matches.append(m.end(0))

        # Search right-to-left for the longest match:
        matches.reverse()
        namespace_prefix: str
        suffix: str
        match_end: int
        for match_end in matches:
            namespace_prefix = body[:match_end]
            suffix = body[match_end:]
            if namespace_prefix in self.namespace_prefixes:
                # We have a winner.
                namespace_id = self.namespace_prefixes[namespace_prefix]

                if self.output_only_used_namespaces:
                    self.used_namespaces.add(namespace_id)

                # Return the namespace-prefixed URI:
                return namespace_id + ":" + suffix, True

        # Take the longest possible section, which is now first in the list:
        if len(matches) > 0:
            match_end = matches[0]
            namespace_prefix = body[:match_end]
            suffix = body[match_end:]
        else:
            # Take the entire body:
            namespace_prefix = body
            suffix = ""

        # Build a non-colliding namespace ID.
        while True:
            namespace_id = self.namespace_id_prefix
            if self.namespace_id_use_uuid:
                namespace_id += self.local_namespace_uuid + "-"
            namespace_id += str(self.namespace_id_counter).zfill(self.namespace_id_zfill)
            self.namespace_id_counter += 1
            if namespace_id not in self.namespace_ids:
                # Save the namespace ID for later reuse:
                self.namespace_ids[namespace_id] = namespace_prefix
                self.namespace_prefixes[namespace_prefix] = namespace_id

                if self.output_only_used_namespaces:
                    self.used_namespaces.add(namespace_id)

                # Return the namespace-prefixed URI:
                return namespace_id + ":" + suffix, True

    def escape_pipe(self, item: str)->str:
        # ensure that vertical bars (pipes) are escaped.
        if self.escape_pipes:
            return item.replace('|', '\\|')
        else:
            return item

    def convert_string(self, item: str, line_number: int)->typing.Tuple[str, bool]:
        # Convert this to a KGTK string.
        #
        # Our parser guarantees that double quoted strings use proper
        # escapes... except for vertical bars (pipes).  We have extra work to do to
        # ensure that vertical bars (pipes) are escaped.
        return self.escape_pipe(item), True
 
    def generate_new_node_symbol(self)->str:
        new_node_symbol: str = self.newnode_prefix
        if self.newnode_use_uuid:
            new_node_symbol += self.local_namespace_uuid + "-"
        new_node_symbol += str(self.newnode_counter).zfill(self.newnode_zfill)
        self.newnode_counter += 1
        return new_node_symbol
    
    def convert_boolean(self, item: str, value: str, line_number: int)->typing.Tuple[str, bool]:
        if value == 'true' or value == '1':
            return KgtkFormat.TRUE_SYMBOL, True
        elif value == 'false' or value == '0':
            return KgtkFormat.FALSE_SYMBOL, True
        else:
            if self.verbose:
                print("Line %d: invalid boolean item '%s'>" % (line_number, item), file=self.error_file, flush=True)
            return item, False
        

    def convert_structured_literal(self, item: str, line_number: int, ew: KgtkWriter)->typing.Tuple[str, bool]:
        # This is the subset of strictured literals that fits the
        # pattern "STRING"^^<URI>.

        # Start by splitting on '^^'. We are certain it exists, and that the rightmost
        # instance is the one we want.
        uparrows: int = item.rfind("^^")
        if uparrows < 0:
            # This shouldn't happen!
            if self.verbose:
                print("Line %d: no uparrows in '%s'." % (line_number, item), file=self.error_file, flush=True)
            return item, False

        string: str = item[:uparrows]
        uri: str = item[uparrows+2:]

        if uri == '<http://www.w3.org/2001/XMLSchema#string>':
            # Convert this to a KGTK string.
            return self.escape_pipe(string), True
        elif uri == '<http://www.w3.org/2001/XMLSchema#int>':
            # Convert this to a KGTK number:
            return string[1:-1], True
        elif uri == '<http://www.w3.org/2001/XMLSchema#double>':
            # Convert this to a KGTK number:
            return string[1:-1], True
        elif uri == '<http://www.w3.org/2001/XMLSchema#float>':
            # Convert this to a KGTK number:
            return string[1:-1], True
        elif uri == '<http://www.w3.org/2001/XMLSchema#decimal>':
            # Convert this to a KGTK number:
            return string[1:-1], True
        elif uri == '<http://www.w3.org/2001/XMLSchema#boolean>':
            # Convert this to a KGTK boolean:
            return self.convert_boolean(item, string[1:-1], line_number)

        elif uri == '<http://www.w3.org/2001/XMLSchema#dateTime>':
            # Convert this to a KGTK date-and-time:
            #
            # Note: the W3C XML Schema standard allows the now obsolete
            # end-of-day time "24:00:00".
            return '^' + string[1:-1], True

        # TODO: the "date" schema
        # Problem:  it allows timezone offsets after dates without times!

        converted_uri: str
        valid: bool
        converted_uri, valid = self.convert_uri(uri, line_number)
        if not valid:
            return item, False

        new_node_symbol: str = self.generate_new_node_symbol()
        self.write_row(ew, new_node_symbol, self.structured_value_label, string)
        self.write_row(ew, new_node_symbol, self.structured_uri_label, converted_uri)

        return new_node_symbol, True

    def convert_numeric(self, item: str, line_number: int, ew: KgtkWriter)->typing.Tuple[str, bool]:
        return item, True

    def convert(self, item: str, line_number: int, ew: KgtkWriter)->typing.Tuple[str, bool]:
        """
        Convert an ntriples item to KGTK format.

        TODO: update output_line_count for row written here.
        """
        if item.startswith("_:"):
            return self.convert_blank_node(item)
        elif item.startswith("<") and item.endswith(">"):
            return self.convert_uri(item, line_number)
        elif item.startswith('"') and item.endswith('"'):
            return self.convert_string(item, line_number)
        elif item.startswith('"') and item.endswith(">"):
            return self.convert_structured_literal(item, line_number, ew)
        elif item[0] in "+-0123456789.":
            return self.convert_numeric(item, line_number, ew)

        if self.verbose:
            print("Line %d: unrecognized item '%s'" %(line_number, item), file=self.error_file, flush=True)

        return item, False
    
    def convert_and_validate(self, item: str, line_number: int, ew: KgtkWriter)->typing.Tuple[str, bool]:
        result: str
        is_ok: bool
        result, is_ok = self.convert(item, line_number, ew)

        # Just a little bit of paranoia here regarding tabs and ends-of-lines:
        #
        # TODO: perform these checks (and repairs!) in KgtkValue.
        if "\t" in result:
            result = result.replace("\t", "\\t")
        if "\n" in result:
            result = result.replace("\n", "\\n")
        if "\r" in result:
            result = result.replace("\r", "\\r")

        if is_ok and self.validate:
            kv: KgtkValue = KgtkValue(result, options=self.value_options)
            if not kv.validate():
                if self.verbose:
                    print("Input line %d: imported value '%s' (from '%s') is invalid." % (line_number, result, item),
                          file=self.error_file, flush=True)
                return result, False
        return result, True
            

    def get_default_namespaces(self)->int:
        namespace_id: str = "xml-schema-type"
        namespace_prefix: str = "http://www.w3.org/2001/XMLSchema#"
        self.namespace_ids[namespace_id] = namespace_prefix
        self.namespace_prefixes[namespace_prefix] = namespace_id
        return 1
    
    def get_initial_namespaces(self)->int:
        # Read the namespaces.  If no file, use a limited internal
        # default.
        if self.namespace_file_path is None:
            return self.get_default_namespaces()

        if self.verbose:
            print("Processing namespace file file %s" % str(self.reject_file_path), file=self.error_file, flush=True)

        kr: KgtkReader =  KgtkReader.open(self.namespace_file_path,
                                          mode=KgtkReaderMode.EDGE,
                                          options=self.reader_options,
                                          error_file=self.error_file,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,
        )
        namespace_line_count: int = 0
        namespace_row: typing.List[str]
        for namespace_row in kr:
            namespace_line_count += 1
            if namespace_row[kr.label_column_idx] == self.prefix_expansion_label:
                namespace_id: str = namespace_row[kr.node1_column_idx]
                namespace_prefix: str = namespace_row[kr.node2_column_idx]
                if not (namespace_prefix.startswith('"') and namespace_prefix.endswith('"')):
                    if self.verbose:
                        print("The namespace prefix must be a KGKT string: '%s'" % namespace_prefix, file=self.error_file, flush=True)
                    continue
                
                # Strip the delimiting double quotes from the KGTk string.
                # Per RFC 3986, internal double quotes are not allowed in
                # a URL unless percent-encoded, so we needen't bother looking
                # for them.
                namespace_prefix = namespace_prefix[1:-1]

                if namespace_prefix in self.namespace_prefixes:
                    if self.verbose:
                        print("Duplicate initial namespace prefix '%s'" % namespace_prefix, file=self.error_file, flush=True)
                else:
                    self.namespace_prefixes[namespace_prefix] = namespace_id
                if namespace_id in self.namespace_ids:
                    if self.verbose:
                        print("Duplicate initial namespace id '%s'" % namespace_id, file=self.error_file, flush=True)
                else:
                    self.namespace_ids[namespace_id] = namespace_prefix
            else:
                if self.verbose:
                    print("Ignoring initial namespace label '%s'" % namespace_row[kr.label_column_idx], file=self.error_file, flush=True)

        return namespace_line_count
        
    def write_namespaces_to_output(self, ew: KgtkWriter, count_records: bool = True):
        # Append the namespaces to the output file.
        namespace_id: str
        for namespace_id in sorted(self.namespace_ids.keys()):
            if self.output_only_used_namespaces:
                if namespace_id in self.used_namespaces:
                    self.write_row(ew, namespace_id, self.prefix_expansion_label, '"' + self.namespace_ids[namespace_id] + '"')
            else:
                self.write_row(ew, namespace_id, self.prefix_expansion_label, '"' + self.namespace_ids[namespace_id] + '"')

    def write_updated_namespace_file(self):
        # Is there an updated namespaces file?
        if self.updated_namespace_file_path is None:
            return

        if self.verbose:
            print("Opening updated namespaces file %s" % str(self.updated_namespace_file_path), file=self.error_file, flush=True)
        # Open the updated namespaces file.
        un: KgtkWriter = KgtkWriter.open(self.COLUMN_NAMES,
                                         self.updated_namespace_file_path,
                                         mode=KgtkWriter.Mode.EDGE,
                                         require_all_columns=True,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)
        namespace_id: str
        for namespace_id in sorted(self.namespace_ids.keys()):
            un.write([namespace_id, self.prefix_expansion_label, '"' + self.namespace_ids[namespace_id] + '"'])
        un.close()


    def save_namespaces(self, ew: KgtkWriter):
        self.write_namespaces_to_output(ew)
        self.write_updated_namespace_file()

    def parse(self, line: str, line_number: int)->typing.Tuple[typing.List[str], bool]:
        m: typing.Optional[typing.Match] = self.ROW_RE.match(line)
        if m is None:
            if self.verbose:
                print("Line %d: not parsed.\n%s" % (line_number, line), file=self.error_file, flush=True)
            return [ ], False
        return [m.group("node1"), m.group("label"), m.group("node2")], True


    def process(self):
        output_column_names: typing.List[str]
        if self.build_id and self.idbuilder_options is not None:
            self.idbuilder = KgtkIdBuilder.from_column_names(self.COLUMN_NAMES, self.idbuilder_options)
            output_column_names = self.idbuilder.column_names
        else:
            output_column_names = self.COLUMN_NAMES

        if self.verbose:
            print("Opening output file %s" % str(self.output_file_path), file=self.error_file, flush=True)
        # Open the output file.
        ew: KgtkWriter = KgtkWriter.open(output_column_names,
                                         self.output_file_path,
                                         mode=KgtkWriter.Mode.EDGE,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)

        rw: typing.Optional[typing.TextIO] = None
        if self.reject_file_path is not None:
            if self.verbose:
                print("Opening reject file %s" % str(self.reject_file_path), file=self.error_file, flush=True)
            # Open the reject file. Since the input data is not in KGTK format,
            # we use an ordinary file here.
            if str(self.reject_file_path) == "-":
                rw = sys.stdout
            else:
                rw = open(self.reject_file_path, "wt")


        total_input_line_count: int = 0
        reject_line_count: int = 0
        
        namespace_line_count: int = self.get_initial_namespaces()
            
        input_file_path: str
        for input_file_path in self.input_file_paths:
            input_line_count: int = 0
            
            if self.local_namespace_use_uuid or self.namespace_id_use_uuid or self.newnode_use_uuid:
                if self.override_uuid is not None:
                    self.local_namespace_uuid = self.override_uuid # for debugging
                else:
                    # Generate a new local namespace UUID.
                    self.local_namespace_uuid = shortuuid.uuid()

            # Open the input file.
            if self.verbose:
                print("Opening the input file: %s" % input_file_path, file=self.error_file, flush=True)
            infile: typing.TextIO
            if str(input_file_path) == "-":
                infile = sys.stdin
            else:
                infile = open(input_file_path, 'rt') 

            line: str
            for line in infile:
                input_line_count += 1
                total_input_line_count += 1

                row: typing.List[str]
                valid: bool
                row, valid = self.parse(line, input_line_count)
                if not valid:
                    if rw is not None:
                        rw.write(line)
                    reject_line_count += 1
                    continue

                node1: str
                ok_1: bool
                node1, ok_1 = self.convert_and_validate(row[0], input_line_count, ew)

                label: str
                ok_2: bool
                label, ok_2 = self.convert_and_validate(row[1], input_line_count, ew)

                node2: str
                ok_3: bool
                node2, ok_3 = self.convert_and_validate(row[2], input_line_count, ew)

                if ok_1 and ok_2 and ok_3:
                    self.write_row(ew, node1, label, node2)
                else:
                    if rw is not None:
                        rw.write(line)
                    reject_line_count += 1

            if input_file_path != "-":
                infile.close()


                self.save_namespaces(ew)

        if self.verbose:
            print("Processed %d known namespaces." % (namespace_line_count), file=self.error_file, flush=True)
            print("Processed %d records." % (total_input_line_count), file=self.error_file, flush=True)
            print("Rejected %d records." % (reject_line_count), file=self.error_file, flush=True)
            print("Wrote %d records." % (self.output_line_count), file=self.error_file, flush=True)
        
        if ew is not None:
            ew.close()
            
        if rw is not None and self.reject_file_path is not None and self.reject_file_path != "-":
            rw.close()
            
    @classmethod
    def add_arguments(cls, parser: ArgumentParser):

        parser.add_argument(      "--namespace-id-prefix", dest="namespace_id_prefix",
                                  help="The prefix used to generate new namespaces. (default=%(default)s).",
                                  default=cls.DEFAULT_NAMESPACE_ID_PREFIX)
    
        parser.add_argument(      "--namespace-id-use-uuid", dest="namespace_id_use_uuid",
                                  help="Use the local namespace UUID when generating namespaces. " +
                                  "When there are multiple input files, each input file gets its own UUID. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_NAMESPACE_ID_USE_UUID)

        parser.add_argument(      "--namespace-id-counter", dest="namespace_id_counter",
                                  help="The counter used to generate new namespaces. (default=%(default)s).",
                                  type=int, default=cls.DEFAULT_NAMESPACE_ID_COUNTER)
    
        parser.add_argument(      "--namespace-id-zfill", dest="namespace_id_zfill",
                                  help="The width of the counter used to generate new namespaces. (default=%(default)s).",
                                  type=int, default=cls.DEFAULT_NAMESPACE_ID_ZFILL)
    
        parser.add_argument(      "--output-only-used-namespaces", dest="output_only_used_namespaces",
                                  help="Write only used namespaces to the output file. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_OUTPUT_ONLY_USED_NAMESPACES)

        parser.add_argument(      "--allow-lax-uri", dest="allow_lax_uri",
                                  help="Allow URIs that don't begin with a http:// or https://. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_LAX_URI)

        parser.add_argument(      "--local-namespace-prefix", dest="local_namespace_prefix",
                                  help="The namespace prefix for blank nodes. (default=%(default)s).",
                                  default=cls.DEFAULT_LOCAL_NAMESPACE_PREFIX)

        parser.add_argument(      "--local-namespace-use-uuid", dest="local_namespace_use_uuid",
                                  help="Generate a UUID for the local namespace. " +
                                  "When there are multiple input files, each input file gets its own UUID. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_LOCAL_NAMESPACE_USE_UUID)

        parser.add_argument(      "--prefix-expansion-label", dest="prefix_expansion_label",
                                  help="The label for prefix expansions in the namespace file. (default=%(default)s).",
                                  default=cls.DEFAULT_PREFIX_EXPANSION)
    
        parser.add_argument(      "--structured-value-label", dest="structured_value_label",
                                  help="The label for value records for ntriple structured literals. (default=%(default)s).",
                                  default=cls.DEFAULT_STRUCTURED_VALUE)
    
        parser.add_argument(      "--structured-uri-label", dest="structured_uri_label",
                                  help="The label for URI records for ntriple structured literals. (default=%(default)s).",
                                  default=cls.DEFAULT_STRUCTURED_URI)
    
        parser.add_argument(      "--newnode-prefix", dest="newnode_prefix",
                                  help="The prefix used to generate new nodes for ntriple structured literals. (default=%(default)s).",
                                  default=cls.DEFAULT_NEWNODE_PREFIX)
    
        parser.add_argument(      "--newnode-use-uuid", dest="newnode_use_uuid",
                                  help="Use the local namespace UUID when generating new nodes for ntriple structured literals. " +
                                  "When there are multiple input files, each input file gets its own UUID. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_NEWNODE_USE_UUID)

        parser.add_argument(      "--newnode-counter", dest="newnode_counter",
                                  help="The counter used to generate new nodes for ntriple structured literals. (default=%(default)s).",
                                  type=int, default=cls.DEFAULT_NEWNODE_COUNTER)
    
        parser.add_argument(      "--newnode-zfill", dest="newnode_zfill",
                                  help="The width of the counter used to generate new nodes for ntriple structured literals. (default=%(default)s).",
                                  type=int, default=cls.DEFAULT_NEWNODE_ZFILL)
    
        parser.add_argument(      "--build-id", dest="build_id",
                                  help="Build id values in an id column. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_BUILD_ID)

        parser.add_argument(      "--escape-pipes", dest="escape_pipes",
                                  help="When true, input pipe characters (|) need to be escaped (\\|) per KGTK file format. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ESCAPE_PIPES)

        parser.add_argument(      "--validate", dest="validate",
                                  help="When true, validate that the result fields are good KGTK file format. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_VALIDATE)

        parser.add_argument(      "--override-uuid", dest="override_uuid",
                                  help="When specified, override UUID generation for debugging. (default=%(default)s).",
                                  default=None)

def main():
    """
    Test the KGTK ntriples importer.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument("-i", "--input-files", dest="input_file_paths", nargs='*',
                        help="The file(s) with the input ntriples data. (default=%(default)s)", type=Path, default="-")

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--reject-file", dest="reject_file_path", help="The KGTK file into which to write rejected records. (default=%(default)s).",
                              type=Path, default=None)
    
    parser.add_argument(      "--namespace-file", dest="namespace_file_path", help="The KGTK file with known namespaces. (default=%(default)s).",
                              type=Path, default=None)
    
    parser.add_argument(      "--updated-namespace-file", dest="updated_namespace_file_path",
                              help="An updated KGTK file with known namespaces. (default=%(default)s).",
                              type=Path, default=None)
    

    KgtkNtriples.add_arguments(parser)
    KgtkIdBuilderOptions.add_arguments(parser)
    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_args(args)
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        print("--input-files %s" % " ".join([str(path) for  path in input_file_paths]), file=error_file, flush=True)
        print("--output-file=%s" % str(args.output_file_path), file=error_file, flush=True)
        # TODO: show ifempty-specific options.
        if args.reject_file_path is not None:
            print("--reject-file=%s" % str(args.reject_file_path), file=error_file, flush=True)
        if args.namespace_file_path is not None:
            print("--namespace-file=%s" % str(args.namespace_file_path), file=error_file, flush=True)
        if args.updated_namespace_file_path is not None:
            print("--updated-namespace-file=%s" % str(args.updated_namespace_file_path), file=error_file, flush=True)
        print("--namespace-id-prefix %s" % args.namespace_id_prefix, file=error_file, flush=True)
        print("--namespace-id-use-uuid %s" % str(args.namespace_id_use_uuid), file=error_file, flush=True)
        print("--namespace-id-counter %s" % str(args.namespace_id_counter), file=error_file, flush=True)
        print("--namespace-id-zfill %s" % str(args.namespace_id_zfill), file=error_file, flush=True)
        print("--output-only-used-namespaces %s" % str(args.output_only_used_namespaces), file=error_file, flush=True)
        print("--allow-lax-uri %s" % str(args.allow_lax_uri), file=error_file, flush=True)
        print("--local-namespace-prefix %s" % args.local_namespace_prefix, file=error_file, flush=True)
        print("--local-namespace-use-uuid %s" % str(args.local_namespace_use_uuid), file=error_file, flush=True)
        print("--prefix-expansion-label %s" % args.prefix_expansion_label, file=error_file, flush=True)
        print("--structured-value-label %s" % args.structured_value_label, file=error_file, flush=True)
        print("--structured-uri-label %s" % args.structured_uri_label, file=error_file, flush=True)
        print("--newnode-prefix %s" % args.newnode_prefix, file=error_file, flush=True)
        print("--newnode-use-uuid %s" % str(args.newnode_use_uuid), file=error_file, flush=True)
        print("--newnode-counter %s" % str(args.newnode_counter), file=error_file, flush=True)
        print("--newnode-zfill %s" % str(args.newnode_zfill), file=error_file, flush=True)
        print("--build-id=%s" % str(args.build_id), file=error_file, flush=True)
        print("--escape-pipes=%s" % str(args.escape_pipes), file=error_file, flush=True)
        print("--validate=%s" % str(args.validate), file=error_file, flush=True)
        if args.override_uuid is not None:
            print("--override_uuid=%s" % str(args.override_uuid), file=error_file, flush=True)            

        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    kn: KgtkNtriples = KgtkNtriples(
        input_file_paths=args.input_file_paths,
        output_file_path=args.output_file_path,
        reject_file_path=args.reject_file_path,
        namespace_file_path=args.namespace_file_path,
        updated_namespace_file_path=args.updated_namespace_file_path,
        namespace_id_prefix=args.namespace_id_prefix,
        namespace_id_use_uuid=args.namespace_id_use_uuid,
        namespace_id_counter=args.namespace_id_counter,
        namespace_id_zfill=args.namespace_id_zfill,
        output_only_used_namespaces=args.output_only_used_namespaces,
        newnode_prefix=args.newnode_prefix,
        newnode_use_uuid=args.newnode_use_uuid,
        newnode_counter=args.newnode_counter,
        newnode_zfill=args.newnode_zfill,
        allow_lax_uri=args.allow_lax_uri,
        local_namespace_prefix=args.local_namespace_prefix,
        local_namespace_use_uuid=args.local_namespace_use_uuid,
        prefix_expansion_label=args.prefix_expansion_label,
        structured_value_label=args.structured_value_label,
        structured_uri_label=args.structured_uri_label,
        build_id=args.build_id,
        escape_pipes=args.escape_pipes,
        idbuilder_options=idbuilder_options,
        validate=args.validate,
        override_uuid=args.override_uuid,
        reader_options=reader_options,
        value_options=value_options,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose)

    kn.process()
    
if __name__ == "__main__":
    main()



"""
Import FrameNet to KGTK.
"""


import sys
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Import FrameNet into KGTK.' 
    }


def add_arguments(parser):
    """
    Parse arguments
    Args:
            parser (argparse.ArgumentParser)
    """
    parser.add_output_file()

def run(output_file: KGTKFiles):
    # import modules locally

    from typing import Tuple, List, Union, Dict
    import nltk
    import pandas as pd
    nltk.download('framenet_v17')
    from nltk.corpus import framenet as fn
    import re
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkwriter import KgtkWriter

    def edge2KGTK(edge: Tuple[str, str, str]) -> pd.Series:
        """
        Gets the edge as triple of subject, object, predicate and converts the edge to the KGTK format
        Args:
            edge: Tuple[str, str, str]
                input edge
        Returns: pd.Series
            pandas Series with keys according to KGTK format at
            https://docs.google.com/document/d/1fbbqgyX0N2EdxLam6hatfke1R-nZWkoN6M1oB_f4aQo/edit#heading=h.a5nlqev5bmm4
        """
        s, p, o = edge

        def clean(e: str) -> str:
            out = e.split(':')[-1].replace('_', ' ')
            return KgtkFormat.stringify(re.sub("([a-z])([A-Z])", "\g<1> \g<2>", out).strip().lower())

        return pd.Series({
            'node1': s,
            'relation': p,
            'node2': o,
            'node1;label': clean(s),
            'node2;label': clean(o),
            'relation;label': clean(p),
            'relation;dimension': '',
            'source': KgtkFormat.stringify('FN'),
            'sentence': ''
        })

    def load_framenet():
        edges=[]
        for frm in fn.frames():
            # frame-frame relations
            for fe in frm.frameRelations:
                edges=pretty_frame_edge(edges, frm_id(fe.superFrameName),
                               frm_id(fe.subFrameName), ncheck(fe.type.name))

            # lexical units
            for lu in frm.lexUnit.keys():
                edges.append([frm_id(frm.name), 'fn:HasLexicalUnit', lu_format(lu, frm.name)])

            # FE
            for fe in frm.FE.values():
                if isinstance(fe.semType, nltk.corpus.reader.framenet.AttrDict):

                    # Sem type
                    semtype_edge=[fe_id(fe.name),
                                    '/r/IsA', #'fn:HasSemType',
                                   fe_semtype_id(fe.semType.name)]
                    if semtype_edge not in edges:
                        edges.append(semtype_edge)


                    # Root type
                    root_edge=[fe_semtype_id(fe.semType.name),
                                    '/r/IsA', # 'fn:st:RootType'
                                   fe_semtype_id(fe.semType.rootType.name)]
                    if root_edge not in edges:
                        edges.append(root_edge)


                    # Super type
                    super_edge=[fe_semtype_id(fe.semType.name),
                                    '/r/IsA', #'fn:st:SuperType',
                                   fe_semtype_id(fe.semType.superType.name)]
                    if super_edge not in edges:
                        edges.append(super_edge)


                    # Sub type
                    for fesub in fe.semType.subTypes:
                        sub_edge=[fe_semtype_id(fesub.name),
                                    '/r/IsA',
                                    fe_semtype_id(fe.semType.name)]
                        if sub_edge not in edges:
                            edges.append(sub_edge)
                        #edges.append([fe_semtype_id(fe.semType.name),
                        #            'fn:st:SubType',
                        #            fe_semtype_id(fesub.name)])


                # Requires FE
                if isinstance(fe.requiresFE, nltk.corpus.reader.framenet.AttrDict):
                    req_edge=[fe_id(fe.name), '/r/HasPrerequisite', fe_id(fe.requiresFE.name)]
                    if req_edge not in edges:
                        edges.append(req_edge)
                    #edges.append([fe_id(fe.name), 'fn:fe:RequiresFE', fe_id(fe.requiresFE.name)])

                # Excludes FE
                if isinstance(fe.excludesFE, nltk.corpus.reader.framenet.AttrDict):
                    excl_edge=[fe_id(fe.name), '/r/RelatedTo', fe_id(fe.excludesFE.name)]
                    if excl_edge not in edges:
                        edges.append(excl_edge)
                    #edges.append([fe_id(fe.name), 'fn:fe:ExcludesFE', fe_id(fe.excludesFE.name)])

                # HasFrameElement - coreType as edge feature
                hasfe_edge=[frm_id(frm.name),
                            '/r/HasA', #'fn:HasFrameElement',
                            fe_id(fe.name)]
                if hasfe_edge not in edges:
                    edges.append(hasfe_edge)
        return edges

    def nosp(s):
        return s.replace(' ', '_').split('.')[0].lower()


    def lu_format(wn: str, _fn: str = "") -> str:
        return f'fn:lu:{nosp(_fn)}:{nosp(wn)}'


    def frm_id(me: str) -> str:
        return f'fn:{nosp(me)}'


    def fe_id(fe: str) -> str:
        return f'fn:fe:{nosp(fe)}'


    def fe_semtype_id(sem: str) -> str:
        return f'fn:st:{nosp(sem)}'


    def ncheck(e):
        if e is None:
            return ""
        elif isinstance(e, nltk.corpus.reader.framenet.AttrDict):
            return e
        else:
            return e


    def pretty_frame_edge(edges, s, o, p):
        pn = nosp(p)
        if pn == 'inheritance':
            edges.append([o, '/r/IsA', s])
        elif pn == 'using':
            edges.append([s, '/r/UsedFor', o])
        elif pn == 'subframe':
            edges.append([s, '/r/HasSubevent', o])
        elif pn == 'precedes':
            edges.append([o, '/r/HasPrerequisite', s])
        elif pn == 'perspective_on':
            edges.append([o, '/r/IsA', s])
        else:
            name = {
                'is_inchoative_of': '/r/Causes',
                'inchoative_of': '/r/Causes',
                #
                'is_causative_of': '/r/Causes',
                'causative_of': '/r/Causes',
                #
                'see_also': '/r/RelatedTo',
                #
                'reframing_mapping': '/r/SimilarTo',
                'metaphor': '/r/SimilarTo',
            }[pn]
            edges.append([s, name, o])
        return edges

    try:
        edges=load_framenet()
        out_columns=['node1', 'relation', 'node2', 'node1;label', 'node2;label','relation;label', 'relation;dimension', 'source', 'sentence']

        output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
        ew: KgtkWriter = KgtkWriter.open(out_columns,
                                         output_kgtk_file,
                                         #mode=input_kr.mode,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         #verbose=self.verbose,
                                         #very_verbose=self.very_verbose
                                         )
        df_ = pd.DataFrame(map(edge2KGTK, edges))

        for i, row in df_.iterrows():
            ew.write(row)

        ew.close()

    except Exception as e:
            kgtk_exception_auto_handler(e)



# TODO: Conver this code to use KgtkFormat and KgtkWriter in various places.
import io
import os
import re
import math
import sys
import time
import redis
from pathlib import Path
import pickle
import typing
import logging
import hashlib
import numpy as np
import pandas as pd  # type: ignore
from tqdm import tqdm  # type: ignore
from ast import literal_eval
from pyrallel import ParallelProcessor
from kgtk.exceptions import KGTKException
from collections import defaultdict, OrderedDict
from sentence_transformers import SentenceTransformer
from SPARQLWrapper import SPARQLWrapper, JSON, POST, URLENCODED  # type: ignore
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.value.kgtkvalueoptions import KgtkValueOptions
from kgtk.kgtkformat import KgtkFormat


class EmbeddingVector:
    def __init__(self, model_name=None, query_server=None, cache_config: dict = None, parallel_count=1):
        self._logger = logging.getLogger(__name__)
        if not model_name:
            self.model_name = 'bert-base-nli-mean-tokens'
        else:
            self.model_name = model_name
        self._logger.info("Using model {}".format(self.model_name))
        self.model = SentenceTransformer(self.model_name)
        # setup redis cache server
        if query_server is None or query_server == "":
            self.wikidata_server = "https://query.wikidata.org/sparql"
        else:
            self.wikidata_server = query_server
        if cache_config and cache_config.get("use_cache", False):
            host = cache_config.get("host", "dsbox01.isi.edu")
            port = cache_config.get("port", 6379)
            self.redis_server = connect_to_redis(host, port)
        else:
            self.redis_server = None
        self._parallel_count = int(parallel_count)
        self._logger.debug("Running with {} processes.".format(parallel_count))
        self.vectors_map = dict()
        self.node_labels = dict()  # this is used to store {node:label} pairs
        self.candidates = defaultdict(dict)  # this is used to store all node {node:dict()} information
        self.vectors_2D = None
        self.vector_dump_file: typing.Optional[str] = None
        self.gt_nodes = set()
        self.metadata = []
        self.gt_indexes = set()
        self.input_format = ""
        self.token_pattern = re.compile(r"(?u)\b\w\w+\b")

    def get_sentences_embedding(self, sentences: typing.List[str], qnodes: typing.List[str]):
        """
            transform a list of sentences to embedding vectors
        """

        if self.redis_server is not None:
            sentence_embeddings = []
            for each_node, each_sentence in zip(qnodes, sentences):
                query_cache_key = each_node + each_sentence
                if self.model_name != "bert-base-wikipedia-sections-mean-tokens":
                    query_cache_key += self.model_name
                cache_res = self.redis_server.get(query_cache_key)
                if cache_res is not None:
                    sentence_embeddings.append(literal_eval(cache_res.decode("utf-8")))
                    # self._logger.error("{} hit!".format(each_node+each_sentence))
                else:
                    each_embedding = self.model.encode([each_sentence], show_progress_bar=False)
                    sentence_embeddings.extend(each_embedding)
                    self.redis_server.set(query_cache_key, str(each_embedding[0].tolist()))
        else:
            sentence_embeddings = self.model.encode(sentences, show_progress_bar=False)
        return sentence_embeddings

    def send_sparql_query(self, query_body: str):
        """
            a simple wrap to send the query and return the returned results
        """
        qm = SPARQLWrapper(self.wikidata_server)
        qm.setReturnFormat(JSON)
        qm.setMethod(POST)
        qm.setRequestMethod(URLENCODED)
        self._logger.debug("Sent query is:")
        self._logger.debug(str(query_body))
        qm.setQuery(query_body)
        try:
            results = qm.query().convert()['results']['bindings']
            return results
        except Exception as e:
            error_message = ("Sending Sparql query to {} failed!".format(self.wikidata_server))
            self._logger.error(error_message)
            self._logger.debug(e, exc_info=True)
            raise KGTKException(error_message)

    def _get_labels(self, nodes: typing.List[str]):
        nodes_need_query = set()
        for each in nodes:
            if each not in self.node_labels:
                nodes_need_query.add(each)
        if nodes_need_query:
            query_nodes = " ".join(["wd:{}".format(each) for each in nodes_need_query])
            query = """
            select ?item ?nodeLabel
            where { 
              values ?item {""" + query_nodes + """}
              ?item rdfs:label ?nodeLabel.
              FILTER(LANG(?nodeLabel) = "en").
            }
            """
            results2 = self.send_sparql_query(query)
            for each_res in results2:
                node_id = each_res['item']['value'].split("/")[-1]
                nodes_need_query.remove(node_id)
                value = each_res['nodeLabel']['value']
                self.node_labels[node_id] = value

            # for those nodes we can't find label, just add this to dict to prevent query again
            for each_node in nodes_need_query:
                self.node_labels[each_node] = each_node

    def _get_labels_and_descriptions(self, query_qnodes: str, need_find_label: bool, need_find_description: bool):
        query_body = """
            select ?item ?itemDescription ?itemLabel
            where {
              values ?item {""" + query_qnodes + """ }
                 SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
            }
        """
        results = self.send_sparql_query(query_body)
        for each in results:
            each_node = each['item']['value'].split("/")[-1]
            if 'itemDescription' in each:
                description = each['itemDescription']['value']
            else:
                description = ""
            if "itemLabel" in each:
                label = each['itemLabel']['value']
            else:
                label = ""
            if need_find_label:
                self.candidates[each_node]["label_properties"] = [label]
            if need_find_description:
                self.candidates[each_node]["description_properties"] = [description]

    def _get_property_values(self, query_qnodes: str, properties: dict, properties_reversed: dict):
        """
        run sparql query to get corresponding property values of given q nodes
        """
        used_p_node_ids = set()
        all_needed_properties = ""
        for part_name, part in properties.items():
            if part_name == "isa_properties":
                self._get_labels(part)

        for each_node, role in properties_reversed.items():
            if role != {"has_properties"} and each_node not in {"label", "description", "all"}:
                all_needed_properties += "wdt:{} ".format(each_node)

        query_body = """
        select ?item ?properties ?eachPropertyValueLabel
        where {{
          values ?item {{{all_nodes}}}
          values ?properties {{{properties}}}
          ?item ?properties ?eachPropertyValue.
          SERVICE wikibase:label {{ bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }}
        }}
        """.format(all_nodes=query_qnodes, properties=all_needed_properties)
        results = self.send_sparql_query(query_body)

        for each_res in results:
            node_id = each_res['item']['value'].split("/")[-1]
            node_property = each_res['properties']['value'].split("/")[-1]
            roles = properties_reversed[node_property]
            value = each_res['eachPropertyValueLabel']['value']
            if node_property in properties["isa_properties"] and self.node_labels[node_property].endswith("of"):
                value = self.node_labels[node_property] + "||" + value
            used_p_node_ids.add(node_property)
            for each_role in roles:
                if each_role != "property_values":
                    if each_role in self.candidates[node_id]:
                        self.candidates[node_id][each_role].add(value)
                    else:
                        self.candidates[node_id][each_role] = {value}
        return used_p_node_ids

    def _get_all_properties(self, query_qnodes: str, used_p_node_ids: set, properties: dict):
        """
        run sparql query to get all properties of given q nodes
        """
        has_properties_set = set(properties["has_properties"])
        query_body3 = """
            select DISTINCT ?item ?p_entity ?p_entityLabel
            where {
              values ?item {""" + query_qnodes + """}
              ?item ?p ?o.
              FILTER regex(str(?p), "^http://www.wikidata.org/prop/P", "i")
              BIND (IRI(REPLACE(STR(?p), "http://www.wikidata.org/prop", "http://www.wikidata.org/entity")) AS ?p_entity) .
              SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
            }
            """
        results3 = self.send_sparql_query(query_body3)
        for each in results3:
            node_name = each['item']['value'].split("/")[-1]
            p_node_id = each['p_entity']['value'].split("/")[-1]
            p_node_label = each['p_entityLabel']['value']
            if p_node_id not in used_p_node_ids:
                if has_properties_set == {"all"} or p_node_id in has_properties_set:
                    if "has_properties" in self.candidates[node_name]:
                        self.candidates[node_name]["has_properties"].add(p_node_label)
                    else:
                        self.candidates[node_name]["has_properties"] = {p_node_label}

    def get_item_description(self, target_properties: dict, properties_reversed: dict,
                             qnodes: typing.Union[set, typing.List[str]]):
        """
            use sparql query to get the descriptions of given Q nodes
        """
        # find_all_properties = False
        if "all" in properties_reversed:
            # find_all_properties = True
            _ = properties_reversed.pop("all")
        self._logger.info("Need to find all properties.")

        hash_generator = hashlib.md5()
        # sort to ensure the hash key same
        target_properties = OrderedDict(sorted(target_properties.items()))
        hash_generator.update(str(target_properties).encode('utf-8'))
        properties_list_hash = "||" + str(hash_generator.hexdigest())

        sentences_cache_dict = {}
        if self.redis_server is not None:
            for each_node in qnodes:
                cache_key = each_node + properties_list_hash
                cache_res = self.redis_server.get(cache_key)
                self._logger.debug("Cached key is: {}".format(cache_key))
                if cache_res is not None:
                    self._logger.debug("Cache hit {}".format(cache_key))
                    sentences_cache_dict[each_node] = cache_res.decode("utf-8")

        self._logger.debug("Cached for those nodes {} / {}".format(len(sentences_cache_dict), len(qnodes)))
        self._logger.debug(str(set(sentences_cache_dict.keys())))
        self._logger.debug(
            "Need run query for those nodes {} / {}:".format(len(qnodes) - len(sentences_cache_dict), len(qnodes)))

        # we do not need to get those node again
        if len(sentences_cache_dict) > 0:
            qnodes = set(qnodes) - set(sentences_cache_dict.keys())
        self._logger.debug(str(qnodes))

        # only need to do query when we still have remained nodes
        if len(qnodes) > 0:
            query_qnodes = ""
            for each in qnodes:
                query_qnodes += "wd:{} ".format(each)

            need_find_label = "label" in target_properties["label_properties"]
            need_find_description = "description" in target_properties["description_properties"]
            # this is used to get corresponding labels / descriptions
            if need_find_label or need_find_description:
                self._get_labels_and_descriptions(query_qnodes, need_find_label, need_find_description)

            # this is used to get corresponding labels of properties values
            used_p_node_ids = self._get_property_values(query_qnodes, target_properties, properties_reversed)

            # if need get all properties, we need to run extra query
            # if find_all_properties:
            self._get_all_properties(query_qnodes, used_p_node_ids, target_properties)

        for each_node_id in qnodes:
            each_sentence = self.attribute_to_sentence(self.candidates[each_node_id], each_node_id)
            self.candidates[each_node_id]["sentence"] = each_sentence
            # add to cache
            if self.redis_server is not None:
                response = self.redis_server.set(each_node_id + properties_list_hash, each_sentence)
                if response:
                    self._logger.debug("Pushed cache for {} success.".format(each_node_id + properties_list_hash))

        for each_node_id, sentence in sentences_cache_dict.items():
            self.candidates[each_node_id]["sentence"] = sentence

    def _process_one(self, args):
        """
        one process for multiprocess calling, should not be used for any other function
        :param args: args to receive from main process
        :return: corresponding node vector and attribute
        """
        node_id = args["node_id"]
        each_node_attributes = args["attribute"]
        concat_sentence = self.attribute_to_sentence(each_node_attributes, node_id)
        vectors = self.get_sentences_embedding([concat_sentence], [node_id])[0]
        return {"v_" + node_id: vectors, "c_" + node_id: each_node_attributes}

    def _multiprocess_collector(self, data):
        for k, v in data.items():
            if k.startswith("v_"):
                k = k.replace("v_", "")
                self.vectors_map[k] = v
            else:
                k = k.replace("c_", "")
                self.candidates[k] = v

    def read_input(self,
                   input_file_path: Path,
                   target_properties: dict,
                   property_labels_dict: dict,
                   skip_nodes_set: set = None,
                   input_format: str = "kgtk_format",
                   black_list_set: typing.Optional[set] = None,
                   error_file: typing.TextIO = sys.stderr,
                   reader_options: typing.Optional[KgtkReaderOptions] = None,
                   value_options: typing.Optional[KgtkValueOptions] = None,
                   verbose: bool = False,
                   ):
        """
            load the input candidates files
        """
        self.node_labels.update(property_labels_dict)
        # reverse sentence property to be {property : role)
        properties_reversed = defaultdict(set)
        pp: typing.Optional[ParallelProcessor] = None

        # This type union becomes painful to work with, below.
        each_node_attributes: typing.Optional[typing.Mapping[str, typing.Union[typing.List, typing.Set]]] = None

        current_process_node_id = None
        node_id = None

        for k, v in target_properties.items():
            for each_property in v:
                properties_reversed[each_property].add(k)

        if input_format == "test_format":
            self.input_format = input_format
            input_file = open(input_file_path, "r") if str(input_file_path) != "-" else sys.stdin
            input_df = pd.read_csv(input_file, dtype=object)
            gt = {}
            count = 0
            if "GT_kg_id" in input_df.columns:
                gt_column_id = "GT_kg_id"
            elif "kg_id" in input_df.columns:
                gt_column_id = "kg_id"
            else:
                raise KGTKException(
                    "Can't find ground truth id column! It should either named as `GT_kg_id` or `kg_id`")

            for _, each in input_df.iterrows():
                temp = []
                if isinstance(each["candidates"], str):
                    temp = str(each['candidates']).split("|")
                elif each['candidates'] is np.nan or math.isnan(each['candidates']):
                    temp = []

                to_remove_q = set()
                if each[gt_column_id] is np.nan:
                    self._logger.warning("Ignore NaN gt value form {}".format(str(each)))
                    each[gt_column_id] = ""
                gt_nodes = each[gt_column_id].split(" ")
                label = str(each["label"])
                if len(gt_nodes) == 0:
                    self._logger.error("Skip a row with no ground truth node given: as {}".format(str(each)))
                    continue
                if label == "":
                    self._logger.error("Skip a row with no label given: as {}".format(str(each)))
                    continue
                temp.extend(gt_nodes)

                for each_q in temp:
                    if skip_nodes_set is not None and each_q in skip_nodes_set:
                        to_remove_q.add(each_q)
                temp = list(set(temp) - to_remove_q)
                count += len(temp)
                self.gt_nodes.add(each[gt_column_id])
                self.get_item_description(target_properties, properties_reversed, temp)

            self._logger.info("Totally {} rows with {} candidates loaded.".format(str(len(gt)), str(count)))

        elif input_format == "kgtk_format":
            # assume the input edge file is sorted
            if "all" in properties_reversed:
                _ = properties_reversed.pop("all")
                add_all_properties = True
            else:
                add_all_properties = False

            self.input_format = input_format

            kr: KgtkReader = KgtkReader.open(input_file_path,
                                             error_file=error_file,
                                             options=reader_options,
                                             value_options=value_options,
                                             verbose=verbose,
                                             )
            if kr.node1_column_idx < 0:
                raise KGTKException("Missing column: node1 or alias")
            if kr.label_column_idx < 0:
                raise KGTKException("Missing column: label or alias")
            if kr.node2_column_idx < 0:
                raise KGTKException("Missing column: node2 or alias")

            self._logger.debug("node1 column index = {}".format(kr.node1_column_idx))
            self._logger.debug("label column index = {}".format(kr.label_column_idx))
            self._logger.debug("node2 column index = {}".format(kr.node2_column_idx))

            # read contents
            each_node_attributes = {
                "has_properties": set(),
                "isa_properties": set(),
                "label_properties": [],
                "description_properties": [],
                "has_properties_values": [],
            }

            if self._parallel_count > 1:
                # need to set with spawn mode to initialize with multiple cuda in multiprocess
                from multiprocessing import set_start_method
                set_start_method('spawn')
                pp = ParallelProcessor(self._parallel_count, self._process_one, collector=self._multiprocess_collector)
                pp.start()

            row: typing.List[str]
            for row in kr:
                node_id = row[kr.node1_column_idx]
                # skip nodes id in black list
                if black_list_set and node_id in black_list_set:
                    continue

                node_property = row[kr.label_column_idx]
                node_value = row[kr.node2_column_idx]

                # CMR: the following code looks like it was intended to remove
                # any language code and language suffix.  It would have the
                # side effect of removing location coordinates entirely.
                #
                # remove @ mark
                if "@" in node_value and node_value[0] != "@":
                    node_value = node_value[:node_value.index("@")]

                # in case we meet an empty value, skip it
                if node_value == "":
                    self._logger.warning("""Skip line "{}" because of empty value.""".format(row))
                    continue

                # CMR: Better to use KgtkFormat.unstringify(node_value), as it will remove escapes from
                # internal double or single quotes.
                #
                # remove extra double quote " and single quote '
                while len(node_value) >= 3 and node_value[0] == '"' and node_value[-1] == '"':
                    node_value = node_value[1:-1]
                while len(node_value) >= 3 and node_value[0] == "'" and node_value[-1] == "'":
                    node_value = node_value[1:-1]

                if current_process_node_id != node_id:
                    if current_process_node_id is None:
                        current_process_node_id = node_id
                    else:
                        # if we get to next id, concat all properties into one sentence to represent the Q node

                        current_process_node_id, each_node_attributes = self.process_qnode(current_process_node_id,
                                                                                           each_node_attributes,
                                                                                           node_id, pp)

                if node_property in properties_reversed:
                    roles = properties_reversed[node_property].copy()
                    node_value = self.get_real_label_name(node_value)
                    # if we get property_values, it should be saved to isa-properties part
                    if "property_values" in roles:
                        # for property values part, changed to be "{property} {value}"
                        node_value_combine = self.get_real_label_name(node_property) + " " + self.get_real_label_name(node_value)
                        if each_node_attributes is None:
                            raise ValueError("each_node_attributes is missing")
                        if not isinstance(each_node_attributes["has_properties_values"], list):
                            raise ValueError('each_node_attributes["has_properties_values"] is not a list.')
                        each_node_attributes["has_properties_values"].append(node_value_combine)
                        # remove those 2 roles in case we have duplicate using of this node later
                        roles.discard("property_values")
                        roles.discard("has_properties")
                    for each_role in roles:
                        attrs: typing.Union[typing.List, typing.Set] = each_node_attributes[each_role]
                        if isinstance(attrs, set):
                            attrs.add(node_value)
                        elif isinstance(attrs, list):
                            attrs.append(node_value)
                        else:
                            raise ValueError('each_node_attributes[%s] is not a list or set.' % repr(each_role))

                elif add_all_properties:  # add remained properties if need all properties
                    attrs2: typing.Union[typing.List, typing.Set] = each_node_attributes["has_properties"]
                    if isinstance(attrs2, list):
                        attrs2.append(self.get_real_label_name(node_property))
                    else:
                        raise ValueError('each_node_attributes["has_properties"] is not a list.')

                # close multiprocess pool
                if self._parallel_count > 1 and pp is not None:
                    pp.task_done()
                    pp.join()
        else:
            raise KGTKException("Unknown input format {}".format(input_format))

        # case where there was a single qnode in the input file
        unprocessed_qnode = False
        if each_node_attributes:
            for k in each_node_attributes:
                if each_node_attributes[k]:
                    unprocessed_qnode = True
                    break
        if unprocessed_qnode:
            a, b = self.process_qnode(current_process_node_id, each_node_attributes, node_id, pp)
        self._logger.info("Totally {} Q nodes loaded.".format(len(self.candidates)))
        try:
            file_path = input_file_path.name
            file_name = file_path[:file_path.rfind(".")]
        except AttributeError:
            file_name = "input_from_memory"
        self.vector_dump_file = "dump_vectors_{}_{}.pkl".format(file_name, self.model_name)
        # self._logger.debug("The cache file name will be {}".format(self.vector_dump_file))

    def process_qnode(self, current_process_node_id, each_node_attributes, node_id, pp):
        # for multi process
        if self._parallel_count > 1:
            each_arg = {"node_id": current_process_node_id, "attribute": each_node_attributes}
            pp.add_task(each_arg)
        # for single process
        else:
            concat_sentence = self.attribute_to_sentence(each_node_attributes, current_process_node_id)
            each_node_attributes["sentence"] = concat_sentence
            self.candidates[current_process_node_id] = each_node_attributes
        # after write down finish, we can clear and start parsing next one
        each_node_attributes = {"has_properties": set(), "isa_properties": set(), "label_properties": [],
                                "description_properties": [], "has_properties_values": []}
        # update to new id
        current_process_node_id = node_id
        return current_process_node_id, each_node_attributes

    def get_real_label_name(self, node: str)->str:
        if node in self.node_labels:
            return self.node_labels[node].replace('"', "")
        else:
            return node

    def attribute_to_sentence(self, attribute_dict: dict, node_id=None):
        concated_sentence = ""
        have_isa_properties = False

        # sort the properties to ensure the sentence always same
        attribute_dict = {key: sorted(list(value)) for key, value in attribute_dict.items() if len(value) > 0}
        if "label_properties" in attribute_dict and len(attribute_dict["label_properties"]) > 0:
            concated_sentence += self.get_real_label_name(attribute_dict["label_properties"][0])
        if "description_properties" in attribute_dict and len(attribute_dict["description_properties"]) > 0:
            if concated_sentence != "" and attribute_dict["description_properties"][0] != "":
                concated_sentence += ", "
            concated_sentence += self.get_real_label_name(attribute_dict["description_properties"][0])
        if "isa_properties" in attribute_dict and len(attribute_dict["isa_properties"]) > 0:
            have_isa_properties = True
            temp_str: str = ""
            for each in attribute_dict["isa_properties"]:
                each = self.get_real_label_name(each)
                if "||" in each:
                    if "instance of" in each:
                        each = each.split("||")[1]
                    else:
                        each = each.replace("||", " ")
                temp_str += each + ", "
            if concated_sentence != "" and temp_str != "":
                concated_sentence += " is "
            elif concated_sentence == "":
                concated_sentence += "It is "
            # remove last ", "
            concated_sentence += temp_str[:-2]
        if "has_properties_values" in attribute_dict and len(attribute_dict["has_properties_values"]) > 0:
            temp_list: typing.List[str] = [self.get_real_label_name(each) for each in attribute_dict["has_properties_values"]]
            if concated_sentence != "":
                if not have_isa_properties:
                    concated_sentence += " "
                else:
                    concated_sentence += ", "
            else:
                concated_sentence += "It "
            concated_sentence += " and ".join(temp_list)
        if "has_properties" in attribute_dict and len(attribute_dict["has_properties"]) > 0:
            temp_list2: typing.List[str] = [self.get_real_label_name(each) for each in attribute_dict["has_properties"]]
            temp_list2 = list(set(temp_list2))
            if concated_sentence != "" and temp_list2[0] != "":
                if have_isa_properties:
                    concated_sentence += ", and has "
                else:
                    concated_sentence += " has "
            elif temp_list2[0] != "":
                concated_sentence += "It has "
            concated_sentence += " and ".join(temp_list2)
        # add ending period
        if concated_sentence != "":
            concated_sentence += "."
        self._logger.debug("Transform node {} --> {}".format(node_id, concated_sentence))
        return concated_sentence

    def get_vectors(self):
        """
            main function to get the vector representations of the descriptions
        """
        if self._parallel_count == 1:
            start_all = time.time()
            self._logger.info("Now generating embedding vector.")
            for q_node, each_item in tqdm(self.candidates.items()):
                # do process for each row(one target)
                sentence = each_item["sentence"]
                if isinstance(sentence, bytes):
                    sentence = sentence.decode("utf-8")
                vectors = self.get_sentences_embedding([sentence], [q_node])
                self.vectors_map[q_node] = vectors[0]
            self._logger.info("Totally used {} seconds.".format(str(time.time() - start_all)))
        else:
            # Skip get vector function because we already get them
            pass

    def dump_vectors(self, file_name, type_=None):
        if file_name.endswith(".pkl"):
            file_name = file_name.replace(".pkl", "")
        if type_ == "2D":
            with open(file_name + ".pkl", "wb") as f:
                pickle.dump(self.vectors_2D, f)
            dimension = len(self.vectors_2D[0])
            # TODO: This should probably be converted to KgtkWriter.
            with open(file_name + ".tsv", "w") as f:
                for each in self.vectors_2D:
                    for i, each_val in enumerate(each):
                        _ = f.write(str(each_val))
                        if i != dimension - 1:
                            _ = f.write("\t")
                    _ = f.write("\n")
        elif type_ == "metadata":
            with open(file_name + "_metadata.tsv", "w") as f:
                for each in self.metadata:
                    _ = f.write(each + "\n")
        else:
            with open(file_name + ".pkl", "wb") as f:
                pickle.dump(self.vectors_map, f)
            # TODO: This should probably be converted to KgtkWriter.
            with open(file_name + ".tsv", "w") as f:
                for each in self.vectors_map.values():
                    for i in each:
                        _ = f.write(str(i) + "\t")
                    _ = f.write("\n")

    def print_vector(self, vectors, output_properties: str = "text_embedding",
                     output_format="kgtk_format", save_embedding_sentence=False):
        self._logger.debug("START printing the vectors")
        if output_format == "kgtk_format":
            # TODO: This should be comverted to use KgtkWriter
            print("node\tproperty\tvalue\n", end="")
            all_nodes = list(self.vectors_map.keys())
            ten_percent_len = math.ceil(len(vectors) / 10)
            for i, each_vector in enumerate(vectors):
                if i % ten_percent_len == 0:
                    percent = i / ten_percent_len * 10
                    self._logger.debug("Finished {}%".format(percent))
                print("{}\t{}\t".format(all_nodes[i], output_properties), end="")
                for each_dimension in each_vector[:-1]:
                    print(str(each_dimension) + ",", end="")
                print(str(each_vector[-1]))
                if save_embedding_sentence:
                    print("{}\t{}\t{}".format(all_nodes[i], "embedding_sentence",
                                              self.candidates[all_nodes[i]]["sentence"]))

        elif output_format == "tsv_format":
            for each_vector in vectors:
                for each_dimension in each_vector[:-1]:
                    print(str(each_dimension) + "\t", end="")
                print(str(each_vector[-1]))
        self._logger.debug("END printing the vectors")

    def plot_result(self, output_properties: dict, input_format="kgtk_format",
                    output_uri: str = "", output_format="kgtk_format",
                    dimensional_reduction="none", dimension_val=2,
                    save_embedding_sentence=False
                    ):
        """
            transfer the vectors to lower dimension so that we can plot
            Then save the 2D vector file for further purpose
        """
        self.vectors_map = {k: v for k, v in sorted(self.vectors_map.items(), key=lambda item: item[0], reverse=True)}
        vectors = list(self.vectors_map.values())
        # reduce dimension if needed
        if dimensional_reduction.lower() == "tsne":
            self._logger.warning("Start running TSNE to reduce dimension. It will take some time.")
            start = time.time()
            from sklearn.manifold import TSNE  # type: ignore
            self.vectors_2D = TSNE(n_components=int(dimension_val), random_state=0).fit_transform(vectors)
            self._logger.info("Totally used {} seconds.".format(time.time() - start))
        elif dimensional_reduction.lower() == "pca":
            self._logger.warning("Start running PCA to reduce dimension. It will take some time.")
            start = time.time()
            from sklearn.decomposition import PCA  # type: ignore
            self.vectors_2D = PCA(n_components=int(dimension_val)).fit_transform(vectors)
            self._logger.info("Totally used {} seconds.".format(time.time() - start))
        elif dimensional_reduction.lower() == "none":
            self._logger.info("Not run dimensional reduction algorithm.")
        else:
            raise KGTKException("Unknown or unsupport dimensional reduction type: {}".format(dimensional_reduction))

        if output_uri not in {"", "none"}:
            if not os.path.exists(output_uri):
                raise ValueError("The given metadata output folder does not exist!")

            if self.vector_dump_file is None:
                raise ValueError("vector_dump_file is None")

            metadata_output_path = os.path.join(output_uri, self.vector_dump_file.split("/")[-1])
            if input_format == "test_format":
                gt_indexes = set()
                vector_map_keys = list(self.vectors_map.keys())
                for each_node in self.gt_nodes:
                    gt_indexes.add(vector_map_keys.index(each_node))

                self.metadata.append("Q_nodes\tType\tLabel\tDescription")
                for i, each in enumerate(self.vectors_map.keys()):
                    label = self.node_labels[each]
                    description = self.candidates[each]["sentence"]
                    if i in gt_indexes:
                        self.metadata.append("{}\tground_truth_node\t{}\t{}".format(each, label, description))
                    else:
                        self.metadata.append("{}\tcandidates\t{}\t{}".format(each, label, description))
                self.gt_indexes = gt_indexes

            elif input_format == "kgtk_format":
                if len(output_properties.get("metadata_properties", [])) == 0:
                    for k, v in self.candidates.items():
                        label = v.get("label_properties", "")
                        if len(label) > 0 and isinstance(label, list):
                            label = label[0]
                        description = v.get("description_properties", "")
                        if len(description) > 0 and isinstance(description, list):
                            description = description[0]
                        self.metadata.append("{}\t\t{}\t{}".format(k, label, description))
                else:
                    required_properties = output_properties["metadata_properties"]
                    self.metadata.append("node\t" + "\t".join(required_properties))
                    for k, v in self.candidates.items():
                        each_metadata = k + "\t"
                        for each in required_properties:
                            each_metadata += v.get(each, " ") + "\t"
                        self.metadata.append(each_metadata)
            self.dump_vectors(metadata_output_path, "metadata")

        output_props: str = output_properties.get("output_properties", "text_embedding")
        if self.vectors_2D is not None:
            self.print_vector(self.vectors_2D,
                              output_props,
                              output_format,
                              save_embedding_sentence)
        else:
            self.print_vector(vectors,
                              output_props,
                              output_format,
                              save_embedding_sentence)

    def evaluate_result(self):
        """
        for the ground truth nodes, evaluate the average distance to the centroid, the lower the average distance,
        the better clustering results should be
        """
        centroid = None
        gt_nodes_vectors = []
        if len(self.gt_indexes) == 0:
            points = set(range(len(self.vectors_map)))
        else:
            points = self.gt_indexes
        for i, each in enumerate(self.vectors_map.keys()):
            if i in points:
                if centroid is None:
                    centroid = np.array(self.vectors_map[each])
                else:
                    centroid += np.array(self.vectors_map[each])
                gt_nodes_vectors.append(self.vectors_map[each])
        centroid = centroid / len(points)

        distance_sum = 0
        for each in gt_nodes_vectors:
            distance_sum += self.calculate_distance(each, centroid)
        self._logger.info(
            "The average distance for the ground truth nodes to centroid is {}".format(distance_sum / len(points)))

    @staticmethod
    def calculate_distance(a, b):
        if len(a) != len(b):
            raise KGTKException("Vector dimension are different!")
        dist = 0
        for v1, v2 in zip(a, b):
            dist += (v1 - v2) ** 2
        dist = dist ** 0.5
        return dist


def connect_to_redis(host, port):
    _logger = logging.getLogger(__name__)
    redis_server = redis.Redis(host=host, port=port, db=0)
    try:
        _ = redis_server.get("foo")
        _logger.debug("Cache server {}:{} connected!".format(host, port))
    except Exception as e:
        _logger.error("Cache server {}:{} is not able to be connected! Will not use cache!".format(host, port))
        _logger.debug(e, exc_info=True)
        redis_server = None
    return redis_server



import datetime
import importlib
from io import StringIO
import itertools
import os
import pkgutil
import signal
import sys
import time
import typing

from kgtk import cli
from kgtk.exceptions import KGTKException, KGTKExceptionHandler, KGTKArgumentParseException
from kgtk import __version__
from kgtk.cli_argparse import KGTKArgumentParser, add_shared_arguments, add_default_arguments, CheckDepsAction
import sh # type: ignore


# module name should NOT start with '__' (double underscore)
handlers = [x.name for x in pkgutil.iter_modules(cli.__path__)
                   if not x.name.startswith('__')]

# import signal
# signal.signal(signal.SIGPIPE, signal.SIG_DFL)

pipe_delimiter = '/'
sequential_delimiter = '/,'
parallel_delimiter = '/:'
ret_code = 0

def cmd_done(cmd, success, exit_code):
    # cmd.cmd -> complete command line
    global ret_code
    ret_code = exit_code


_save_progress: bool = False
_save_progress_tty: typing.Optional[str] = None
_save_progress_command: typing.Optional[typing.Any] = None
def progress_startup(pid: typing.Optional[int] = None, fd: typing.Optional[int] = None):
    # This can be called multiple times, if it desired to monitor several
    # input files in sequence.
    #
    # If pid is None, the cirrent process will be monitored.  If pid is not
    # None, the specified process will be monitored.
    #
    # If target_fd is None, them all fds will be monitored.  If target is not
    # None, the specific fd will be monitored: it must already be open, and it
    # should be an input file.  There is no option to moitor multiple specific
    # input files other than calling this routine sequentially
    import os
    import sh
    global _save_progress, _save_progress_tty
    if _save_progress and _save_progress_tty is not None:
        global _save_progress_command
        if _save_progress_command is not None:
            # Shut down an existing process monitor.
            try:
                _save_progress_command.terminate()
            except Exception:
                pass
            _save_progress_command = None

        # Start a process monitor.
        if pid is None:
            pid = os.getpid()
        if fd is None:
            _save_progress_command = sh.pv("-d {}".format(pid), _out=_save_progress_tty, _err=_save_progress_tty, _bg=True)
        else:
            _save_progress_command = sh.pv("-d {}:{}".format(pid, fd),
                                           _out=_save_progress_tty, _err=_save_progress_tty, _bg=True)

def progress_shutdown():
    global _save_progress_command
    if _save_progress_command is not None:
        try:
            _save_progress_command.terminate()
        except Exception:
            pass
        _save_progress_command = None
    
def split_list(sequence, sep):
    """From stack overflow;
    https://stackoverflow.com/questions/54372218/how-to-split-a-list-into-sublists-based-on-a-separator-similar-to-str-split
    """
    chunk = []
    for val in sequence:
        if val == sep:
            yield chunk
            chunk = []
        else:
            chunk.append(val)
    yield chunk

def cli_entry_pipe(args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)->int:
    # parse internal pipe
    import copy
    ret_code: int = 0
    pipe = [list(y) for x, y in itertools.groupby(args, lambda a: a == pipe_delimiter) if not x]
    if len(pipe) == 0:
        parser.print_usage()
        parser.exit(KGTKArgumentParseException.return_code)
    elif len(pipe) == 1:  # single command
        cmd_args = pipe[0]
        cmd_name = cmd_args[0].replace("_", "-")
        cmd_args[0] = cmd_name
        # build sub-parser
        if cmd_name in subparser_lookup:
            mod, sub_parser = subparser_lookup[cmd_name]
            add_default_arguments(sub_parser)  # call this before adding other arguments
            if cmd_name not in subparsers_built:
                if hasattr(mod, 'add_arguments_extended'):
                    psa = copy.deepcopy(parsed_shared_args)
                    psa._command = cmd_name
                    mod.add_arguments_extended(sub_parser, psa)
                else:
                    mod.add_arguments(sub_parser)
                subparsers_built.add(cmd_name)
        parsed_args = parser.parse_args(cmd_args)

        # load module
        kwargs = {}
        func = None
        if parsed_args.cmd:
            h = parsed_args.cmd
            func = mod.run

            # remove sub-command name
            kwargs = vars(parsed_args)
            del kwargs['cmd']

            # set shared arguments
            for sa in vars(parsed_shared_args):
                if sa not in sub_parsers.choices[h].shared_arguments:
                    del kwargs[sa]
                else:
                    kwargs[sa] = getattr(parsed_shared_args, sa)

        global _save_progress
        _save_progress = parsed_shared_args._progress
        global _save_progress_tty
        _save_progress_tty = parsed_shared_args._progress_tty
        if parsed_shared_args._progress:
            if hasattr(mod, 'custom_progress') and mod.custom_progress():
                pass
            else:
                progress_startup()

        # run module
        try: 
          kgtk_exception_handler = KGTKExceptionHandler(debug=parsed_shared_args._debug)
          ret_code = kgtk_exception_handler(func, **kwargs)
        except KeyboardInterrupt as e:
            print("\nKeyboard interrupt in %s." % " ".join(args), file=sys.stderr, flush=True)
            progress_shutdown()
            if hasattr(mod, 'keyboard_interrupt'):
                mod.keyboard_interrupt()

            # Silently exit instead of re-raising the KeyboardInterrupt.
            # raise

    else:  # piped commands
        if parsed_shared_args._pipedebug:
            print("Building a KGTK pipe.  pid=%d" % (os.getpid()), file=sys.stderr, flush=True)
        processes = [ ]
        try:
            for idx, cmd_args in enumerate(pipe):
                # add shared arguments
                full_args = [ ]
                for shared_arg in shared_args:
                    if idx == 0 or str(shared_arg) != "--progress":
                        full_args.append(shared_arg)
                full_args.extend(cmd_args)
                kwargs = {
                    "_bg_exc": False,
                    "_done": cmd_done,
                    "_err": sys.stderr,
                    "_bg": True,
                    "_internal_bufsize": 1,
                }

                # add specific arguments
                if idx == 0:  # The first commamd reads from our STDIN.
                    kwargs["_in"] = sys.stdin

                if idx + 1 < len(pipe):
                    # All commands but the last pipe their output to the next command.
                    kwargs["_piped"] = True
                else:
                    # The last command writes to our STDOUT.
                    kwargs["_out"] = sys.stdout

                if parsed_shared_args._pipedebug:
                    cmd_str = " ".join(full_args)
                    for key in kwargs:
                        cmd_str += " " + key + "=" + str(kwargs[key])
                    print("pipe[%d]: kgtk %s" % (idx, cmd_str), file=sys.stderr, flush=True)

                if idx == 0:
                    processes.append(sh.kgtk(*full_args, **kwargs))
                else:
                    processes.append(sh.kgtk(processes[idx - 1], *full_args, **kwargs))
                
            for idx in range(len(pipe)):
                processes[idx].wait()

        except sh.SignalException_SIGPIPE:
            if parsed_shared_args._pipedebug:
                print("\npipe: sh.SignalException_SIGPIPE", file=sys.stderr, flush=True)

        except sh.SignalException_SIGTERM:
            if parsed_shared_args._pipedebug:
                print("\npipe: sh.SignalException_SIGTERM", file=sys.stderr, flush=True)
            raise

        except KeyboardInterrupt as e:
            if parsed_shared_args._pipedebug:
                print("\npipe: KeyboardInterrupt", file=sys.stderr, flush=True)
            if len(processes) > 0:
                for idx in range(len(processes)):
                    process = processes[idx]
                    pgid = process.pgid
                    print("Killing cmd %d process group %d" % (idx, pgid), file=sys.stderr, flush=True)
                    process.signal_group(signal.SIGINT)

        except sh.ErrorReturnCode as e:
            if parsed_shared_args._pipedebug:
                print("\npipe: sh.ErrorReturnCode", file=sys.stderr, flush=True)
            # mimic parser exit
            parser.exit(KGTKArgumentParseException.return_code, e.stderr.decode('utf-8'))
    
    return ret_code

def cli_entry_sequential_commands(args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)->int:
    # parse internal sequence of pipes
    ret_code: int = 0
    for commands in split_list(args, sequential_delimiter):
        ret_code = cli_entry_pipe(commands, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)
        if ret_code != 0:
            break

    return ret_code

def cli_entry(*args):
    """
    Usage:
        kgtk <command> [options]
    """
    global ret_code

    # Capture the initial time for timing measurements.
    start_time: float = time.time()
    process_start_time: float = time.process_time()

    # get all arguments
    if not args:
        args = tuple(sys.argv)
    args = args[1:]

    # base parser for shared arguments
    base_parser = KGTKArgumentParser(add_help=False)
    base_parser.add_argument(
        '-V', '--version',
        action='version',
        version='KGTK %s' % __version__,
        help='show KGTK version number and exit.'
    )
    base_parser.add_argument(
        '--check-deps',
        action=CheckDepsAction,
        help='check dependencies',
    )
    shared_args = base_parser.add_argument_group('shared optional arguments')
    shared_args.add_argument('--debug', dest='_debug', action='store_true', default=False, help='enable debug mode')
    shared_args.add_argument('--expert', dest='_expert', action='store_true', default=False, help='enable expert mode')
    shared_args.add_argument('--pipedebug', dest='_pipedebug', action='store_true', default=False, help='enable pipe debug mode')
    shared_args.add_argument('--progress', dest='_progress', action='store_true', default=False, help='enable progress monitoring')
    shared_args.add_argument('--progress-tty', dest='_progress_tty', action='store', default="/dev/tty", help='progress monitoring output tty')
    shared_args.add_argument('--timing', dest='_timing', action='store_true', default=False, help='enable timing measurements')
    add_shared_arguments(shared_args)

    # parse shared arguments
    parsed_shared_args, rest_args = base_parser.parse_known_args(args)
    shared_args = tuple(filter(lambda a: a not in rest_args, args))
    args = tuple(rest_args)

    # complete parser, load sub-parser of each module
    parser = KGTKArgumentParser(
        parents=[base_parser], prog='kgtk',
        description='kgtk --- Knowledge Graph Toolkit',
    )
    sub_parsers = parser.add_subparsers(
        metavar='command',
        dest='cmd'
    )
    subparsers_built = set()
    subparser_lookup = {}
    sub_parsers.required = True
    for h in handlers:
        mod = importlib.import_module('.{}'.format(h), 'kgtk.cli')
        subp = mod.parser()
        # only create sub-parser with sub-command name and defer full build
        cmd: str = h.replace("_", "-")
        sub_parser = sub_parsers.add_parser(cmd, **subp)
        subparser_lookup[cmd] = (mod, sub_parser)
        if 'aliases' in subp:
            for alias in subp['aliases']:
                subparser_lookup[alias] = (mod, sub_parser)

    # add root level usage after sub-parsers are created
    # this won't pollute help info in sub-parsers
    parser.usage = '%(prog)s [options] command [ / command]*'

    ret_code = cli_entry_sequential_commands(args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)

    if parsed_shared_args._timing:
        end_time: float = time.time()
        elapsed_seconds: float = end_time - start_time

        process_end_time: float = time.process_time()
        process_elapsed_seconds: float = process_end_time - process_start_time

        cpu_ratio: float = process_elapsed_seconds / elapsed_seconds

        print("Timing: elapsed=%s CPU=%s (%5.1f%%): %s" % (str(datetime.timedelta(seconds=elapsed_seconds)),
                                                           str(datetime.timedelta(seconds=process_elapsed_seconds)),
                                                           cpu_ratio * 100.0,
                                                           " ".join(args)), file=sys.stderr, flush=True)

    return ret_code




"""Copy records from the first KGTK file to the output file,
building one column (usually node2) from discrete subfields.

TODO: Need KgtkWriterOptions
"""

from argparse import _MutuallyExclusiveGroup, Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Copy a KGTK file, building one column (usualy node2) from seperate columns for each subfield.',
        'description': 'Copy a KGTK file, building one column (usually node2) from seperate columns for each subfield. ' +
        '\n\nStrings may include language qualified strings, and quantities may include numbers. ' +
        '\n\nDate and times subfields and symbol subfields may be optionally quoted. Triple quotes may be used where quotes are accepted. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert implode --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalue import KgtkValueFields
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_output_file()
    parser.add_output_file(who="The KGTK file for records that are rejected.",
                           dest="reject_file",
                           options=["--reject-file"],
                           metavar="REJECT_FILE",
                           optional=True)

    parser.add_argument(      "--column", dest="column_name", help="The name of the column to explode. (default=%(default)s).", default=KgtkFormat.NODE2)

    parser.add_argument(      "--prefix", dest="prefix", help="The prefix for exploded column names. (default=%(default)s).",
                              default=KgtkFormat.NODE2 + ";" + KgtkFormat.KGTK_NAMESPACE)

    parser.add_argument(      "--types", dest="type_names", nargs='*',
                               help="The KGTK data types for which fields should be imploded. (default=%(default)s).",
                               choices=KgtkFormat.DataType.choices(),
                               default=KgtkFormat.DataType.choices())

    parser.add_argument(      "--without", dest="without_fields", nargs='*',
                               help="The KGTK fields to do without. (default=%(default)s).",
                               choices=KgtkValueFields.OPTIONAL_DEFAULT_FIELD_NAMES,
                               default=None)

    parser.add_argument(      "--overwrite", dest="overwrite_column", metavar='True/False',
                              help="Indicate that it is OK to overwrite an existing imploded column. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--validate", dest="validate", metavar='True/False',
                              help="Validate imploded values. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--escape-pipes", dest="escape_pipes", metavar='True/False',
                              help="When true, pipe characters (|) need to be escaped (\\|) per KGTK file format. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--quantities-include-numbers", dest="quantities_include_numbers", metavar='True/False',
                              help="When true, numbers are acceptable quantities. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--general-strings", dest="general_strings", metavar='True/False',
                              help="When true, strings may include language qualified strings. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--remove-prefixed-columns", dest="remove_prefixed_columns", metavar='True/False',
                              help="When true, remove all columns beginning with the prefix from the output file. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--ignore-unselected-types", dest="ignore_unselected_types", metavar='True/False',
                              help="When true, input records with valid but unselected data types will be passed through to output. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--retain-unselected-types", dest="retain_unselected_types", metavar='True/False',
                              help="When true, input records with valid but unselected data types will be retain existing data on output. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--build-id", dest="build_id", metavar='True/False',
                              help="Build id values in an id column. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--show-data-types", dest="show_data_types", metavar='True/False',
                              help="Print the list of data types and exit. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--quiet", dest="quiet", metavar='True/False',
                              help="When true, suppress certain complaints unless verbose. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkIdBuilderOptions.add_arguments(parser, expert=_expert)
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        reject_file: KGTKFiles,

        column_name: str,
        prefix: str,
        type_names: typing.List[str],
        without_fields: typing.Optional[typing.List[str]],
        overwrite_column: bool,
        validate: bool,
        escape_pipes: bool,
        quantities_include_numbers: bool,
        general_strings: bool,
        remove_prefixed_columns: bool,
        ignore_unselected_types: bool,
        retain_unselected_types: bool,
        build_id: bool,
        show_data_types: bool,
        quiet: bool,
        
        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.reshape.kgtkimplode import KgtkImplode
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
    reject_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(reject_file, who="KGTK reject file")

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_dict(kwargs)    
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file, flush=True)
        if reject_kgtk_file is not None:
            print("--reject-file=%s" % str(reject_kgtk_file), file=error_file, flush=True)

        print("--column %s" % column_name, file=error_file, flush=True)
        print("--prefix %s" % prefix, file=error_file, flush=True)
        print("--overwrite %s" % str(overwrite_column), file=error_file, flush=True)
        print("--validate %s" % str(validate), file=error_file, flush=True)
        print("--escape-pipes %s" % str(escape_pipes), file=error_file, flush=True)
        print("--quantities-include-numbers %s" % str(quantities_include_numbers), file=error_file, flush=True)
        print("--general-strings %s" % str(general_strings), file=error_file, flush=True)
        print("--remove-prefixed-columns %s" % str(remove_prefixed_columns), file=error_file, flush=True)
        print("--ignore-unselected-types %s" % str(ignore_unselected_types), file=error_file, flush=True)
        print("--retain-unselected-types %s" % str(retain_unselected_types), file=error_file, flush=True)
        if type_names is not None:
            print("--types %s" % " ".join(type_names), file=error_file, flush=True)
        if without_fields is not None:
            print("--without %s" % " ".join(without_fields), file=error_file, flush=True)
        print("--show-data-types %s" % str(show_data_types), file=error_file, flush=True)
        print("--quiet %s" % str(quiet), file=error_file, flush=True)
        print("--build-id=%s" % str(build_id), file=error_file, flush=True)
        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)
    if show_data_types:
        data_type: str
        for data_type in KgtkFormat.DataType.choices():
            print("%s" % data_type, file=error_file, flush=True)
        return 0

    wf: typing.List[str] = without_fields if without_fields is not None else list()

    try:
        ex: KgtkImplode = KgtkImplode(
            input_file_path=input_kgtk_file,
            output_file_path=output_kgtk_file,
            reject_file_path=reject_kgtk_file,
            column_name=column_name,
            prefix=prefix,
            type_names=type_names,
            without_fields=wf,
            overwrite_column=overwrite_column,
            validate=validate,
            escape_pipes=escape_pipes,
            quantities_include_numbers=quantities_include_numbers,
            general_strings=general_strings,
            remove_prefixed_columns=remove_prefixed_columns,
            ignore_unselected_types=ignore_unselected_types,
            retain_unselected_types=retain_unselected_types,
            quiet=quiet,
            build_id=build_id,
            idbuilder_options=idbuilder_options,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose)

        ex.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""Unreify values in a KGTK file.

TODO: Need KgtkWriterOptions
"""

from argparse import _MutuallyExclusiveGroup, Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Unreify values in a KGTK file.',
        'description': 'Read a KGTK file, such as might have been created by importing an ntriples file.  ' +
        'Search for reified values and transform them into an unreified form.' +
        '\n\nAn ID column will be added to the output file if not present in the input file.  ' +
        '\n\n--reified-file PATH, if specified, will get a copy of the input records that were ' +
        'identified as reified values. ' +
        '\n\n--uninvolved-file PATH, if specified, will get a copy of the input records that were ' +
        ' identified as not being reified values. ' +
        '\n\n--unreified-file PATH, if specified, will get a copy of the unreified output records, which ' +
        ' will still be written to the main output file.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert unreify-values --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.unreify.kgtkunreifyvalues import KgtkUnreifyValues
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(who="The KGTK input file with the reified data. " +
                        "It must have node1, label, and node2 columns, or their aliases. " +
                        "It may have an ID column;  if it does not, one will be appended to the output file. " +
                        "It may not have any additional columns. ")

    parser.add_output_file(who="The KGTK file to write output records with unreified data. " +
                        "This file may differ in shape from the input file by the addition of an ID column. " +
                        "The records in the output file will not, generally, be in the same order as they appeared in the input file. ")
    
    parser.add_output_file(who="A KGTK output file that will contain only the reified RDF statements.",
                           dest="reified_file",
                           options=["--reified-file"],
                           metavar="REIFIED_FILE",
                           optional=True)
    parser.add_output_file(who="A KGTK output file that will contain only the unreified RDF statements.",
                           dest="unreified_file",
                           options=["--unreified-file"],
                           metavar="UNREIFIED_FILE",
                           optional=True)
    parser.add_output_file(who="A KGTK output file that will contain only the uninvolved input.",
                           dest="uninvolved_file",
                           options=["--uninvolved-file"],
                           metavar="UNINVOLVED_FILE",
                           optional=True)
    
    KgtkUnreifyValues.add_arguments(parser)
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        reified_file: KGTKFiles,
        unreified_file: KGTKFiles,
        uninvolved_file: KGTKFiles,

        trigger_label_value: str,
        trigger_node2_value: str,
        value_label_value: str,
        old_label_value: str,
        new_label_value: typing.Optional[str],

        allow_multiple_values: bool,
        allow_extra_columns: bool,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.unreify.kgtkunreifyvalues import KgtkUnreifyValues
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
    reified_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(reified_file, who="KGTK reified file")
    unreified_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(unreified_file, who="KGTK unreified file")
    uninvolved_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(uninvolved_file, who="KGTK uninvolved file")

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-files %s" % str(input_kgtk_file), file=error_file, flush=True)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file, flush=True)
        if reified_kgtk_file is not None:
            print("--reified-file=%s" % str(reified_kgtk_file), file=error_file, flush=True)
        if unreified_kgtk_file is not None:
            print("--unreified-file=%s" % str(unreified_kgtk_file), file=error_file, flush=True)
        if uninvolved_kgtk_file is not None:
            print("--uninvolved-file=%s" % str(uninvolved_kgtk_file), file=error_file, flush=True)
        
        print("--trigger-label=%s" % trigger_label_value, file=error_file, flush=True)
        print("--trigger-node2=%s" % trigger_node2_value, file=error_file, flush=True)
        print("--value-label=%s" % value_label_value, file=error_file, flush=True)
        print("--old-label=%s" % old_label_value, file=error_file, flush=True)
        if new_label_value is not None:
            print("--new-label=%s" % new_label_value, file=error_file, flush=True)

        print("--allow-multiple-values=%s" % str(allow_multiple_values), file=error_file, flush=True)
        print("--allow-extra-columns=%s" % str(allow_extra_columns), file=error_file, flush=True)

        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        kuv: KgtkUnreifyValues = KgtkUnreifyValues(
            input_file_path=input_kgtk_file,
            output_file_path=output_kgtk_file,
            reified_file_path=reified_kgtk_file,
            unreified_file_path=unreified_kgtk_file,
            uninvolved_file_path=uninvolved_kgtk_file,

            trigger_label_value=trigger_label_value,
            trigger_node2_value=trigger_node2_value,
            value_label_value=value_label_value,
            old_label_value=old_label_value,
            new_label_value=new_label_value,

            allow_multiple_values=allow_multiple_values,
            allow_extra_columns=allow_extra_columns,

            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose)

        kuv.process()
    
        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))

